@@ -168,7 +168,8 @@
 #define UDM_LOCK_SQL            3
 #define UDM_LOCK_SEGMENTER      4
 #define UDM_LOCK_DB             5
-#define UDM_LOCK_MAX		6
+#define UDM_LOCK_LOG		6
+#define UDM_LOCK_MAX		7
 
 /************************ Statistics **********************/
 typedef struct stat_struct {
@@ -185,21 +186,39 @@ typedef struct stat_list_struct{
 /************************ VARLISTs ************************/
 
 /* Various variable flags */
-#define UDM_VARFLAG_NOCLONE	1	/* Ignore for clone detection */ 
-#define UDM_VARFLAG_USERDEF	2	/* User defined section       */
+#define UDM_VARFLAG_NOCLONE	1  /* Ignore for clone detection        */ 
+#define UDM_VARFLAG_USERDEF	2  /* User defined section              */
+#define UDM_VARFLAG_KEEPVAL	4  /* If value doesn't need to be freed */
+
+/* Variable types */
+#define UDM_VAR_INT       1
+#define UDM_VAR_STR       2
+#define UDM_VAR_DOUBLE    3
+#define UDM_VAR_ENV       16
+#define UDM_VAR_AGENT     17
+#define UDM_VAR_RESULT    18
+#define UDM_VAR_DOCUMENT  19
+#define UDM_VAR_SQLRESULT 20
+
+#define UDM_VARSRC_QSTRING 1
+#define UDM_VARSRC_ENV     2
+
+struct udm_var_handler_st;
 
 typedef struct udm_var_st {
-	int			section;	/**< Number 0..255   */
-	size_t			maxlen;		/**< Max length      */
-	size_t			curlen;		/**< Cur length      */
-	char			*val;		/**< Field Value     */
-	char			*name;		/**< Field Name      */
-	int			flags;
+	struct udm_var_handler_st *handler; /* Variable handler */
+	int		section;	/**< Number 0..255   */
+	size_t		maxlen;		/**< Max length      */
+	size_t		curlen;		/**< Cur length      */
+	char		*val;		/**< Field Value     */
+	char		*name;		/**< Field Name      */
+	int		flags;
 } UDM_VAR;
 
 typedef struct udm_varlist_st {
 	int		freeme;
 	size_t		nvars;
+	size_t		mvars;
 	UDM_VAR		*Var;
 } UDM_VARLIST;
 
@@ -330,6 +349,7 @@ typedef struct {
 	int	follow;		/**< Whether to follow links  */
 	int	use_robots;	/**< Whether to use robots.txt*/
 	int	use_clones;	/**< Whether to detect clones */
+        int	doc_per_site;
 } UDM_SPIDERPARAM;
 
 /*****************************************************************/
@@ -341,6 +361,7 @@ typedef struct {
 	char		*pattern;
 	regex_t		*reg;
 	char		*arg;
+	char		*arg1;
 } UDM_MATCH;
 
 typedef struct {
@@ -387,6 +408,7 @@ typedef struct {
         urlid_t site_id;
         urlid_t server_id;
         urlid_t rec_id;
+        size_t  max_doc_per_site;
 } UDM_HREF;
 
 typedef struct {
@@ -731,6 +753,7 @@ typedef struct udm_config_struct {
 	UDM_MATCHLIST	SectionFilters; /**< IndexIf, NoIndexIf         */
         UDM_MATCHLIST   SectionMatch;   /**< User defined sections      */
 	
+	UDM_HREFLIST	Hrefs;		/**< Links cache                */
 	UDM_RESULT	Targets;	/**< Targets cache              */
 	
 	UDM_VARLIST	Sections;	/**< document section to parse  */
@@ -756,7 +779,8 @@ typedef struct udm_config_struct {
 	void (*ThreadInfo)(struct udm_indexer_struct *,const char * state,const char * str);
 	void (*LockProc)(struct udm_indexer_struct *,int command,int type,const char *fname,int lineno);
 	void (*RefInfo)(int code,const char *url, const char *ref);
-
+	int (*ThreadCreate)(void **thd, void *(*start_routine)(void *), void *arg);
+	int (*ThreadJoin)(void *thd);
 #ifdef MECAB
         mecab_t         *mecab;
 #endif
@@ -779,7 +803,6 @@ typedef struct udm_indexer_struct{
 	UDM_ENV		*Conf;		/**< Configuration                */
 	UDM_LANGMAP	*LangMap;	/**< LangMap for current document */
 	UDM_RESULT	Indexed;	/**< Indexed cache              */
-	UDM_HREFLIST	Hrefs;		/**< Links cache                */
 	
 	/** Cache mode limits */
 	UDM_SEARCH_LIMIT	limits[MAX_SEARCH_LIMIT];
@@ -884,7 +907,6 @@ typedef struct {
         void    (*next_e)(void *t);
         const char *e;
         const char *b;
-        const char **lt;
         const char *s;
 	size_t	ntoks;
 	struct {
