@@ -26,6 +26,8 @@ extern __C_LINK UDM_URL * __UDMCALL UdmURLInit(UDM_URL *url);
 extern __C_LINK void __UDMCALL UdmURLFree(UDM_URL *url);
 extern int UdmURLParse(UDM_URL *url,const char *s);
 extern char * UdmURLNormalizePath(char * path);
+extern size_t UdmURLCanonize(const char *src, char *dst, size_t dstsize);
+extern size_t UdmURLCanonizePath(char *dst, size_t dstsize, const char *src);
 
 #endif
 
