@@ -21,17 +21,26 @@
 extern __C_LINK UDM_VARLIST * __UDMCALL UdmVarListInit(UDM_VARLIST * vars);
 extern __C_LINK void __UDMCALL UdmVarListFree(UDM_VARLIST * vars);
 extern void UdmVarFree(UDM_VAR *);
+extern int UdmVarType(UDM_VAR *);
 
 extern int UdmVarListAddEnviron(UDM_VARLIST *vars, const char *name);
 
 extern __C_LINK int __UDMCALL UdmVarListAddStr(UDM_VARLIST * vars,const char * name, const char * val);
+extern int UdmVarListAddQueryStr(UDM_VARLIST * vars,const char * name, const char * val);
 extern int UdmVarListAddInt(UDM_VARLIST * vars,const char * name, int val);
 extern int UdmVarListAddUnsigned(UDM_VARLIST * vars, const char * name, uint4 val);
 extern int UdmVarListAddDouble(UDM_VARLIST * vars, const char * name, double val);
 extern int UdmVarListAdd(UDM_VARLIST * Lst,UDM_VAR * S);
 extern int UdmVarListDel(UDM_VARLIST * Lst,const char *name);
+extern int UdmVarListDelBySection(UDM_VARLIST * Lst, int secno);
+extern int UdmVarListDelByName(UDM_VARLIST * Lst, const char *name);
 extern int UdmVarListReplace(UDM_VARLIST * Lst,UDM_VAR * S);
-
+extern int UdmVarListCreateObject(UDM_VARLIST *Lst,
+                                  const char *name, int type,
+                                  UDM_VAR **args, size_t nargs);
+extern int UdmVarListInvokeMethod(UDM_VARLIST *Lst,
+                                  UDM_VAR *Var, const char *method,
+                                  UDM_VAR **args, size_t nargs);
 extern int UdmVarListAddLst(UDM_VARLIST * Lst,UDM_VARLIST *Src, const char *name, const char *w);
 extern int UdmVarListInsLst(UDM_VARLIST * Lst,UDM_VARLIST *Src, const char *name, const char *w);
 extern int UdmVarListReplaceLst(UDM_VARLIST * Lst,UDM_VARLIST *Src, const char *name, const char *w);
@@ -49,6 +58,7 @@ extern UDM_VAR *UdmVarListFindWithValue(UDM_VARLIST * vars,const char * name,con
 
 extern __C_LINK const char * __UDMCALL UdmVarListFindStr(UDM_VARLIST * vars,const char * name,const char * defval);
 extern __C_LINK int          UdmVarListFindInt(UDM_VARLIST * vars,const char * name,int defval);
+extern __C_LINK int          UdmVarListFindBool(UDM_VARLIST * vars,const char * name,int defval);
 extern __C_LINK unsigned     UdmVarListFindUnsigned(UDM_VARLIST * vars, const char * name, unsigned defval);
 extern __C_LINK double       UdmVarListFindDouble(UDM_VARLIST * vars, const char * name, double defval);
 
