@@ -82,6 +82,7 @@ __C_LINK const char * __UDMCALL UdmFollowStr(int method)
     case UDM_FOLLOW_PATH:  return "Path";
     case UDM_FOLLOW_SITE:  return "Site";
     case UDM_FOLLOW_WORLD: return "World";
+    case UDM_FOLLOW_URLLIST: return "URLList";
   }
   return "<Unknown follow type>";
 }
@@ -96,6 +97,7 @@ int UdmFollowType(const char * follow)
   if(!strcasecmp(follow,"path"))return(UDM_FOLLOW_PATH);
   if(!strcasecmp(follow,"site"))return(UDM_FOLLOW_SITE);
   if(!strcasecmp(follow,"world"))return(UDM_FOLLOW_WORLD);
+  if(!strcasecmp(follow,"urllist"))return(UDM_FOLLOW_URLLIST);
   return(UDM_FOLLOW_UNKNOWN);
 }
 
@@ -326,7 +328,8 @@ static int add_srv(void *Cfg, size_t ac,char **av)
     Href.method=UDM_METHOD_GET;
     Href.site_id = C->Srv->site_id;
     Href.server_id = C->Srv->site_id;
-    UdmHrefListAdd(&Indexer->Hrefs, &Href);
+    Href.hops= (uint4) UdmVarListFindInt(&C->Srv->Vars, "StartHops", 0);
+    UdmHrefListAdd(&Conf->Hrefs, &Href);
   }
   UDM_FREE(C->Srv->Match.pattern);
   UdmVarListDel(&C->Srv->Vars,"AuthBasic");
@@ -587,12 +590,23 @@ static int add_parser(void *Cfg, size_t ac,char **av)
 }
 
 
+static int add_separator(UDM_VARLIST *Vars, const char *name, const char *val)
+{
+  UDM_DSTR buf;
+  UdmDSTRInit(&buf, 128);
+  UdmDSTRReset(&buf);
+  UdmDSTRAppendf(&buf, "separator.%s", name);
+  UdmVarListAddStr(Vars, buf.data, val);
+  UdmDSTRFree(&buf);
+  return UDM_OK;
+}
+
+
 static int add_section(void *Cfg, size_t ac,char **av)
 {
   UDM_CFG  *C=(UDM_CFG*)Cfg;
   UDM_ENV  *Conf=C->Indexer->Conf;
   UDM_VAR  S;
-  UDM_DSTR buf;
   int      cdon;
 
   bzero((void*)&S, sizeof(S));
@@ -618,17 +632,24 @@ static int add_section(void *Cfg, size_t ac,char **av)
       break;
   }
   S.flags= cdon ? 0 : UDM_VARFLAG_NOCLONE;
-  
-  if (ac == 1)
+
+  if (ac == 0)
   {
-    UdmDSTRInit(&buf, 512);
-    UdmDSTRReset(&buf);
-    UdmDSTRAppendf(&buf, "separator.%s", S.name);
-    UdmVarListAddStr(&Conf->Vars, buf.data, av[0]);
-    UdmDSTRFree(&buf);
+    /* no optional arguments */
   }
-  else if (ac == 2)
+  else if (ac == 1)
   {
+    /* <sep> */
+    add_separator(&Conf->Vars, S.name, av[0]);
+  }
+  else if (ac >= 2 && ac <= 4)
+  {
+    /* 
+       <expr> <repl>
+       <sep> <expr> <repl>
+       <sep> <src> <expr> <repl>
+    */
+
     UDM_MATCH M;
     char err[120]= "";
 
@@ -636,8 +657,27 @@ static int add_section(void *Cfg, size_t ac,char **av)
     M.match_type= UDM_MATCH_REGEX;
     M.case_sense= 1;
     M.section= S.name;
-    M.pattern= av[0];
-    M.arg= av[1];
+
+    switch (ac)
+    {
+      case 2:
+        M.pattern= av[0];
+        M.arg= av[1];
+        break;
+
+      case 3:
+        add_separator(&Conf->Vars, S.name, av[0]);
+        M.pattern= av[1];
+        M.arg= av[2];
+        break;
+
+      case 4:
+        add_separator(&Conf->Vars, S.name, av[0]);
+        M.arg1= av[1];
+        M.pattern= av[2];
+        M.arg= av[3];
+        break;
+    }
 
     if(UDM_OK != UdmMatchListAdd(C->Indexer, &Conf->SectionMatch, &M,
                                  err, sizeof(err), ++C->ordre))
@@ -647,6 +687,11 @@ static int add_section(void *Cfg, size_t ac,char **av)
     }
     S.flags+= UDM_VARFLAG_USERDEF;
   }
+  else
+  {
+    sprintf(Conf->errstr,"too many argiments: '%s'", av[0]);
+    return UDM_ERROR;
+  }
 
   UdmVarListReplace(&Conf->Sections,&S);
   return UDM_OK;
@@ -789,7 +834,6 @@ static int add_url(void *Cfg, size_t ac,char **av)
 {
   UDM_CFG  *C=(UDM_CFG*)Cfg;
   UDM_ENV  *Conf=C->Indexer->Conf;
-  UDM_AGENT *Indexer = C->Indexer;
   
   if(C->flags&UDM_FLAG_ADD_SERV)
   {
@@ -801,7 +845,7 @@ static int add_url(void *Cfg, size_t ac,char **av)
       UdmHrefInit(&Href);
       Href.url=av[1];
       Href.method=UDM_METHOD_GET;
-      UdmHrefListAdd(&Indexer->Hrefs, &Href);
+      UdmHrefListAdd(&Conf->Hrefs, &Href);
     }
     UDM_FREE(al);
   }
@@ -863,7 +907,15 @@ static int flush_srv_table(void *Cfg, size_t ac,char **av)
   }
   return res;
 }
-    
+
+static int dblist_free(void *Cfg, size_t ac,char **av)
+{
+  UDM_CFG  *C=(UDM_CFG*)Cfg;
+  UDM_ENV  *Conf=C->Indexer->Conf;
+  UdmDBListFree(&Conf->dbl);
+  return UDM_OK;
+}
+
 static int env_rpl_charset(void *Cfg, size_t ac,char **av)
 {
   UDM_CFG  *C=(UDM_CFG*)Cfg;
@@ -1061,7 +1113,7 @@ static int env_rpl_bool_var(void *Cfg, size_t ac,char **av)
 {
   UDM_CFG  *C=(UDM_CFG*)Cfg;
   UDM_ENV  *Conf=C->Indexer->Conf;
-  int res=!strcasecmp(av[1],"yes");
+  int res= !strcasecmp(av[1],"yes") || atoi(av[1]) == 1;
   if(!strcasecmp(av[0], "CVSIgnore")) Conf->CVS_ignore = res;
   UdmVarListReplaceInt(&Conf->Vars,av[0],res);
   return UDM_OK;
@@ -1070,7 +1122,7 @@ static int env_rpl_bool_var(void *Cfg, size_t ac,char **av)
 static int srv_rpl_bool_var(void *Cfg, size_t ac,char **av)
 {
   UDM_CFG  *C=(UDM_CFG*)Cfg;
-  int res=!strcasecmp(av[1],"yes");
+  int res= !strcasecmp(av[1],"yes") || atoi(av[1]) == 1;
   UdmVarListReplaceInt(&C->Srv->Vars,av[0],res);
   return UDM_OK;
 }
@@ -1152,6 +1204,7 @@ static UDM_CONFCMD commands[] =
   {"Listen",               1,1,   env_rpl_var},
   {"UseRemoteContentType", 1,1,   env_rpl_var},
   {"UseCRC32URLId",        1,1,   env_rpl_var},
+  {"UseCRC32SiteId",       1,1,   env_rpl_var},
   {"NewsExtensions",       1,1,   env_rpl_var},
   {"CrossWords",           1,1,   env_rpl_var},
   {"SyslogFacility",       1,1,   env_rpl_var},
@@ -1181,6 +1234,9 @@ static UDM_CONFCMD commands[] =
   {"HlEnd",                1,1,   env_rpl_var},
   {"Log2stderr",           1,1,   env_rpl_var},
   {"PagesPerScreen",       1,1,   env_rpl_var},
+#ifdef HAVE_SETVBUF
+  {"StdoutBufferSize",     1,1,   env_rpl_var},
+#endif
   {"MinWordLength",        1,1,   env_rpl_num_var},
   {"MaxWordLength",        1,1,   env_rpl_num_var},
   {"ExcerptSize",          1,1,   env_rpl_num_var},
@@ -1190,6 +1246,8 @@ static UDM_CONFCMD commands[] =
   {"CVSIgnore",            1,1,   env_rpl_bool_var},
   {"UseHTDBURLId",         1,1,   env_rpl_bool_var},
   {"Suggest",              1,1,   env_rpl_bool_var},
+  {"IndexTime",            1,1,   env_rpl_bool_var},
+  {"ExcerptStopword",      1,1,   env_rpl_bool_var},
 
   {"LocalCharset",         1,1,   env_rpl_charset},
   {"BrowserCharset",       1,1,   env_rpl_charset},
@@ -1217,6 +1275,8 @@ static UDM_CONFCMD commands[] =
   {"UseRobotsTxtURL",      1,1,   srv_rpl_var},
   {"MaxNetErrors",         1,1,   srv_rpl_num_var},
   {"MaxHops",              1,1,   srv_rpl_num_var},
+  {"StartHops",            1,1,   srv_rpl_num_var},
+  {"MaxDocPerSite",        1,1,   srv_rpl_num_var},
   {"ServerWeight",         1,1,   srv_rpl_num_var},
   {"Robots",               1,1,   srv_rpl_bool_var},
   {"DetectClones",         1,1,   srv_rpl_bool_var},
@@ -1255,7 +1315,7 @@ static UDM_CONFCMD commands[] =
   
   {"AddType",              1,100, add_type},
   {"Mime",                 2,3,   add_parser},
-  {"Section",              3,7,   add_section},
+  {"Section",              3,8,   add_section},
   {"Affix",                3,3,   add_affix},
   {"Spell",                3,3,   add_spell},
   {"StopwordFile",         1,1,   add_stoplist},
@@ -1267,6 +1327,7 @@ static UDM_CONFCMD commands[] =
   {"Limit",                1,1,   add_limit},
   {"ServerTable",          1,1,   add_srv_table},
   {"FlushServerTable",     0,0,   flush_srv_table},
+  {"DBListFree",           0,0,   dblist_free},
   
   {NULL,0,0,0}
 };
@@ -1435,25 +1496,11 @@ __C_LINK  int __UDMCALL UdmEnvLoad(UDM_AGENT *Indexer,
   if(UDM_OK == (rc=EnvLoad(&Cfg,cname)))
   {
     UDM_ENV *Env= Indexer->Conf;
-    if(Env->Spells.nitems && Env->Affixes.nitems)
-    {
-      char *err= Env->errstr;
-      int flags= strcasecmp(UdmVarListFindStr(&Env->Vars,
-                                            "IspellUsePrefixes","no"),"no");
-      flags= flags ? 0 : UDM_SPELL_NOPREFIX;
-      
-      if (UdmSpellListListLoad(&Env->Spells, err, 128) ||
-          UdmAffixListListLoad(&Env->Affixes, flags, err, 128))
-      {
-        rc= UDM_ERROR;
-        goto freeex;
-      }
-    }
     
-    /* Sort synonyms */
-    UdmSynonymListSort(&Indexer->Conf->Synonyms);
-    
-    UdmVarListInsStr(&Indexer->Conf->Vars, "Request.User-Agent", UDM_USER_AGENT);
+    if (UDM_OK != (rc= UdmEnvPrepare(Env)))
+      goto freeex;
+
+    UdmVarListInsStr(&Env->Vars, "Request.User-Agent", UDM_USER_AGENT);
   }
 
 freeex:
