@@ -65,7 +65,7 @@
 #include "udm_searchtool.h"
 #include "udm_result.h"
 #include "udm_log.h"
-
+#include "udm_agent.h"
 #include "udm_proto.h"
 #include "udm_host.h"
 #include "udm_hash.h"
@@ -73,6 +73,7 @@
 #include "udm_services.h"
 #include "udm_xmalloc.h"
 #include "udm_searchcache.h"
+#include "udm_store.h"
 
 #define UDM_THREADINFO(A,s,m)	if(A->Conf->ThreadInfo)A->Conf->ThreadInfo(A,s,m)
 
@@ -192,21 +193,65 @@ __C_LINK int __UDMCALL UdmClearDatabase(UDM_AGENT *A)
 }
 
 
+static int
+UdmRegisterChild(UDM_AGENT *Indexer, UDM_DOCUMENT *Doc)
+{
+  int result= UDM_OK;
+  
+  UDM_VAR    *Sec;
+  const char  *parent=NULL;
+  int    parent_id=0;
+    
+  if((Sec=UdmVarListFind(&Doc->Sections,"Header.References")) && Sec->val)
+  {
+    /* 
+      References contains all message IDs of my predecessors,
+      space separated my direct parent is the last in the list.
+    */
+    if ((parent= strrchr(Sec->val,' ')))
+    {
+      /* parent now points to the space character, skip it */
+      ++parent;
+    }
+    else
+    {
+      /* there is only one entry in references, so this is my parent */
+      parent=Sec->val;
+    }  
+  }
+    
+  /* get parent from database */
+  if (parent && strlen(parent) && strchr(parent,'@'))
+  {
+    UDM_DOCUMENT Msg;
+      
+    UdmDocInit(&Msg);
+    UdmVarListReplaceStr(&Msg.Sections,"Header.Message-ID",parent);
+    result = UdmURLAction(Indexer, &Msg, UDM_URL_ACTION_FINDBYMSG);
+    parent_id = UdmVarListFindInt(&Msg.Sections,"ID",0);
+    UdmVarListReplaceInt(&Doc->Sections,"Header.Parent-ID",parent_id);
+    UdmDocFree(&Msg);
+  }
+  
+  /* Now register me with my parent  */
+  if(parent_id)
+    result = UdmURLAction(Indexer, Doc, UDM_URL_ACTION_REGCHILD);
+  return result;
+}
+
 
 static int DocUpdate(UDM_AGENT * Indexer, UDM_DOCUMENT *Doc)
 {
-  int        result=UDM_OK;
-  const char *c;
+  int        rc= UDM_OK;
   int        status=UdmVarListFindInt(&Doc->Sections,"Status",0);
-  urlid_t    origin_id = 0;
   urlid_t    url_id = (urlid_t)UdmVarListFindInt(&Doc->Sections, "ID", 0);
   time_t     next_index_time;
   char       dbuf[64];
-  int        use_crosswords, use_newsext; 
+  int        use_newsext; 
+  int        action;
 
   UDM_GETLOCK(Indexer, UDM_LOCK_CONF);
-  use_crosswords = !strcasecmp(UdmVarListFindStr(&Indexer->Conf->Vars,"CrossWords","no"),"yes");
-  use_newsext    = !strcasecmp(UdmVarListFindStr(&Indexer->Conf->Vars,"NewsExtensions","no"),"yes");
+  use_newsext= !strcasecmp(UdmVarListFindStr(&Indexer->Conf->Vars,"NewsExtensions","no"),"yes");
   UDM_RELEASELOCK(Indexer, UDM_LOCK_CONF);
   
   /* First of all check that URL must be delated */
@@ -214,24 +259,19 @@ static int DocUpdate(UDM_AGENT * Indexer, UDM_DOCUMENT *Doc)
   if(Doc->method==UDM_METHOD_DISALLOW)
   {
     UdmLog(Indexer,UDM_LOG_ERROR,"Deleting %s", UdmVarListFindStr(&Doc->Sections, "URL", ""));
-    result = UdmURLAction(Indexer, Doc, UDM_URL_ACTION_DELETE);
-    return result;
+    rc= UdmURLAction(Indexer, Doc, UDM_URL_ACTION_DELETE);
+    return rc;
   }
 
   next_index_time=time(NULL)+Doc->Spider.period;
-  UdmTime_t2HttpStr(next_index_time,dbuf);
-  UdmVarListReplaceStr(&Doc->Sections,"Next-Index-Time",dbuf);
-  
   switch(status){
   
   case 0: /* No HTTP code */
     if (Doc->connp.Host != NULL) Doc->connp.Host->net_errors++;
     UdmLog(Indexer,UDM_LOG_ERROR,"No HTTP response status");
     next_index_time=time(NULL)+Doc->Spider.net_error_delay_time;
-    UdmTime_t2HttpStr(next_index_time,dbuf);
-    UdmVarListReplaceStr(&Doc->Sections,"Next-Index-Time",dbuf);
-    result = UdmURLAction(Indexer, Doc, UDM_URL_ACTION_SUPDATE);
-    return result;
+    action= UDM_URL_ACTION_SUPDATE;
+    break;
   
   case UDM_HTTP_STATUS_OK:                            /* 200 */
   case UDM_HTTP_STATUS_PARTIAL_OK:                    /* 206 */
@@ -239,29 +279,59 @@ static int DocUpdate(UDM_AGENT * Indexer, UDM_DOCUMENT *Doc)
     {
       UdmLog(Indexer,UDM_LOG_ERROR,"No Content-type header");
       next_index_time=time(NULL)+Doc->Spider.net_error_delay_time;
-      UdmTime_t2HttpStr(next_index_time,dbuf);
-      UdmVarListReplaceStr(&Doc->Sections,"Next-Index-Time",dbuf);
       UdmVarListReplaceInt(&Doc->Sections,"Status",UDM_HTTP_STATUS_INTERNAL_SERVER_ERROR);
       if (Doc->connp.Host != NULL) Doc->connp.Host->net_errors++;
-      result = UdmURLAction(Indexer, Doc, UDM_URL_ACTION_SUPDATE);
-      return result;
+      action= UDM_URL_ACTION_SUPDATE;
+      break;
     }
     else
     {
-        if (Doc->connp.Host != NULL) Doc->connp.Host->net_errors = 0;
+      if (Doc->connp.Host != NULL) Doc->connp.Host->net_errors = 0;
+
+      /* Check clones */
+      if(Doc->method == UDM_METHOD_GET && Doc->Spider.use_clones)
+      {
+        urlid_t    origin_id = 0;
+        if (UDM_OK != (rc= UdmURLAction(Indexer, Doc, UDM_URL_ACTION_FINDORIG)))
+          return rc;
+        origin_id= (urlid_t)UdmVarListFindInt(&Doc->Sections,"Origin-ID",0);
+  
+        if(origin_id && origin_id != url_id)
+        {
+          if (UdmNeedLog(UDM_LOG_EXTRA))
+            UdmLog(Indexer, UDM_LOG_EXTRA, "Duplicate Document %s with #%d", 
+                   UdmVarListFindStr(&Doc->Sections, "URL", ""), origin_id);
+          action= UDM_URL_ACTION_UPDCLONE;
+          break;
+        }
+      }
+  
+      /* Check that document wasn't modified since last indexing */
+      if ((UdmVarListFindInt(&Doc->Sections,"crc32", 0) != 0) 
+          &&  (UdmVarListFindInt(&Doc->Sections,"crc32old",0)==UdmVarListFindInt(&Doc->Sections,"crc32",0)) 
+          &&  (!(Indexer->flags&UDM_FLAG_REINDEX)))
+      {
+        action= UDM_URL_ACTION_SUPDATE;
+        break;
+      }
+  
+      /* For NEWS extension: get rec_id from my */
+      /* parent out of db (if I have one...)    */
+      if (use_newsext && (UDM_OK != (rc= UdmRegisterChild(Indexer, Doc))))
+        return rc;
+
+      action= UDM_URL_ACTION_LUPDATE;
+      break;
     }
+
+  case UDM_HTTP_STATUS_NOT_MODIFIED:                  /* 304 */
+    action= UDM_URL_ACTION_SUPDATE;
     break;
   
   case UDM_HTTP_STATUS_MULTIPLE_CHOICES:              /* 300 */
   case UDM_HTTP_STATUS_MOVED_PARMANENTLY:             /* 301 */
   case UDM_HTTP_STATUS_MOVED_TEMPORARILY:             /* 302 */
   case UDM_HTTP_STATUS_SEE_OTHER:                     /* 303 */
-  case UDM_HTTP_STATUS_NOT_MODIFIED:                  /* 304 */
-    /* FIXME: check that status is changed and remove words if necessary */
-    result = UdmURLAction(Indexer, Doc, UDM_URL_ACTION_SUPDATE);
-    return result;
-    break;
-  
   case UDM_HTTP_STATUS_USE_PROXY:                     /* 305 */
   case UDM_HTTP_STATUS_BAD_REQUEST:                   /* 400 */
   case UDM_HTTP_STATUS_UNAUTHORIZED:                  /* 401 */
@@ -281,16 +351,15 @@ static int DocUpdate(UDM_AGENT * Indexer, UDM_DOCUMENT *Doc)
   case UDM_HTTP_STATUS_UNSUPPORTED_MEDIA_TYPE:        /* 415 */
   case UDM_HTTP_STATUS_NOT_IMPLEMENTED:               /* 501 */
   case UDM_HTTP_STATUS_BAD_GATEWAY:                   /* 502 */
-  case UDM_HTTP_STATUS_NOT_SUPPORTED:                 /* 505 */
-  
+  case UDM_HTTP_STATUS_NOT_SUPPORTED:                 /* 505 */  
     /*
-      FIXME: remove words from database
-      Check last reffering time to remove when
-      there are no links to this document anymore
-      UdmLog(Indexer,UDM_LOG_EXTRA,"Deleting URL");
+      Remove words and outgoing links from database.
+      Keep "urlinfo" and "url" values for DBMode=blob searches,
+      as well as for cached search results. Information from
+      these two tables will be deleted later, after "HoldBadHref"
     */
-    result = UdmURLAction(Indexer, Doc, UDM_URL_ACTION_SUPDATE);
-    return result;
+    action= UDM_URL_ACTION_DUPDATE;
+    break;
   
   case UDM_HTTP_STATUS_INTERNAL_SERVER_ERROR:         /* 500 */
   case UDM_HTTP_STATUS_SERVICE_UNAVAILABLE:           /* 503 */
@@ -300,120 +369,19 @@ static int DocUpdate(UDM_AGENT * Indexer, UDM_DOCUMENT *Doc)
     /* We'll retry later, maybe host is down */
     if (Doc->connp.Host != NULL) Doc->connp.Host->net_errors++;
     next_index_time=time(NULL)+Doc->Spider.net_error_delay_time;
-    UdmTime_t2HttpStr(next_index_time,dbuf);
-    UdmVarListReplaceStr(&Doc->Sections,"Next-Index-Time",dbuf);
-    result = UdmURLAction(Indexer, Doc, UDM_URL_ACTION_SUPDATE);
-    return result;
+    action= UDM_URL_ACTION_SUPDATE;
+    break;
   
   default: /* Unknown status, retry later */
     UdmLog(Indexer,UDM_LOG_WARN,"HTTP %d We don't yet know how to handle it, skipped",status);
-    result = UdmURLAction(Indexer, Doc, UDM_URL_ACTION_SUPDATE);
-    return result;
+    action= UDM_URL_ACTION_SUPDATE;
+    break;
   }
-  
-  
-  if(Doc->method==UDM_METHOD_GET && Doc->Spider.use_clones)
-  {
-    result = UdmURLAction(Indexer, Doc, UDM_URL_ACTION_FINDORIG);
-    if(result!=UDM_OK)return result;
-    origin_id = (urlid_t)UdmVarListFindInt(&Doc->Sections,"Origin-ID",0);
-  }
-  
-  
-  /* Check clones */
-  if((origin_id)&&(origin_id!=url_id))
-  {
-    if (UdmNeedLog(UDM_LOG_EXTRA))
-      UdmLog(Indexer, UDM_LOG_EXTRA, "Duplicate Document %s with #%d", 
-             UdmVarListFindStr(&Doc->Sections, "URL", ""), origin_id);
-    result = UdmURLAction(Indexer, Doc, UDM_URL_ACTION_DELWORDS);
-    if(use_crosswords)
-    {
-      if(result == UDM_OK) result = UdmURLAction(Indexer, Doc, UDM_URL_ACTION_DELCWORDS);
-    }
-    if(result == UDM_OK) result = UdmURLAction(Indexer, Doc, UDM_URL_ACTION_UPDCLONE);
-    return result;
-  }
-  
-  /* Check that document wasn't modified since last indexing */
-  if ((UdmVarListFindInt(&Doc->Sections,"crc32", 0) != 0) 
-      &&  (UdmVarListFindInt(&Doc->Sections,"crc32old",0)==UdmVarListFindInt(&Doc->Sections,"crc32",0)) 
-      &&  (!(Indexer->flags&UDM_FLAG_REINDEX)))
-  {
-    result = UdmURLAction(Indexer, Doc, UDM_URL_ACTION_SUPDATE);
-    return result;
-  }
-  
-  /* Compose site_id string and calc it's Hash32 */
-  /*
-  sprintf(site_id_str,"%s://%s/",Doc->CurURL.schema,Doc->CurURL.hostinfo);
-  UdmVarListReplaceInt(&Doc->Sections, "Site-ID", UdmStrHash32(site_id_str)); for what this need ?
-  */
-  
-  /* Copy default languages, if not given by server and not guessed */
-  if(!(c=UdmVarListFindStr(&Doc->Sections,"Content-Language",NULL)))
-  {
-    if((c=UdmVarListFindStr(&Doc->Sections,"DefaultLang",NULL)))
-      UdmVarListReplaceStr(&Doc->Sections,"Content-Language",c);
-  }
-  
-  /* For NEWS extension: get rec_id from my */
-  /* parent out of db (if I have one...)    */
-  if(use_newsext)
-  {
-    UDM_VAR    *Sec;
-    const char  *parent=NULL;
-    int    parent_id=0;
-    
-    if((Sec=UdmVarListFind(&Doc->Sections,"Header.References")) && Sec->val)
-    {
-      /* References contains all message IDs of my */
-      /* predecessors, space separated             */
-      /* my direct parent is the last in the list  */
-      if((parent = strrchr(Sec->val,' ')))
-      {
-        /* parent now points to the space */
-        /* character skip it              */
-        ++parent;
-      }
-      else
-      {
-        /* there is only one entry in */
-        /* references, so this is my parent */
-        parent=Sec->val;
-      }  
-    }
-    
-    /* get parent from database */
-    if(parent && strlen(parent) && strchr(parent,'@'))
-    {
-      UDM_DOCUMENT Msg;
-      
-      UdmDocInit(&Msg);
-      UdmVarListReplaceStr(&Msg.Sections,"Header.Message-ID",parent);
-      result = UdmURLAction(Indexer, &Msg, UDM_URL_ACTION_FINDBYMSG);
-      parent_id = UdmVarListFindInt(&Msg.Sections,"ID",0);
-      UdmVarListReplaceInt(&Doc->Sections,"Header.Parent-ID",parent_id);
-      UdmDocFree(&Msg);
-    }
-    
-    /* Now register me with my parent  */
-    if(parent_id) result = UdmURLAction(Indexer, Doc, UDM_URL_ACTION_REGCHILD);
-    
-    if(result!=UDM_OK)return result;
-  }
-  
-  /* Now store words and crosswords */
-  if(UDM_OK != (result = UdmURLAction(Indexer, Doc, UDM_URL_ACTION_INSWORDS)))
-    return result;
-  
-  if(use_crosswords)
-    if(UDM_OK != (result = UdmURLAction(Indexer, Doc, UDM_URL_ACTION_INSCWORDS)))
-      return result;
-  
-  result = UdmURLAction(Indexer, Doc, UDM_URL_ACTION_LUPDATE);
-  
-  return result;
+
+  UdmTime_t2HttpStr(next_index_time,dbuf);
+  UdmVarListReplaceStr(&Doc->Sections,"Next-Index-Time",dbuf);  
+  rc= UdmURLAction(Indexer, Doc, action);
+  return rc;
 }
 
 
@@ -473,11 +441,6 @@ static int UdmDocUpdate(UDM_AGENT *Indexer, UDM_DOCUMENT *Doc)
         return rc;
     }
     
-    /* this should be only if we use blob mode, otherwise built-in wan't work
-    if(UDM_OK != (rc = UdmResAction(Indexer, &Indexer->Indexed, UDM_RES_ACTION_INSWORDS)))
-      return rc;
-    */
-   
     if (Indexer->Indexed.num_rows) UdmResultFree(&Indexer->Indexed);
   }
   else
@@ -587,9 +550,7 @@ __C_LINK int __UDMCALL UdmResAction(UDM_AGENT *A, UDM_RESULT *R, int cmd)
   UDM_DB  *db;
   size_t i, dbfrom = 0, dbto;
   
-  UDM_GETLOCK(A, UDM_LOCK_CONF);
   dbto =  A->Conf->dbl.nitems;
-  UDM_RELEASELOCK(A, UDM_LOCK_CONF);
 
   for (i = dbfrom; i < dbto; i++)
   {
@@ -673,104 +634,159 @@ __C_LINK int __UDMCALL UdmSrvAction(UDM_AGENT *A, UDM_SERVERLIST *S, int cmd)
 
 static const int search_cache_size=1000;
 
+static
+int UdmFindWordsDB(UDM_AGENT *A, UDM_RESULT *CurRes, UDM_DB *db)
+{
+  int rc= UDM_OK;
+  const char *dbaddr= UdmVarListFindStr(&db->Vars,"DBAddr","<noaddr>");
+      
+  UdmLog(A, UDM_LOG_DEBUG, "UdmFind for %s", dbaddr);
+      
+  switch(db->DBDriver)
+    {
+    case UDM_DB_SEARCHD:
+      rc= UdmFindWordsSearchd(A, CurRes, db);
+      break;
+#ifdef HAVE_SQL
+    default:
+      rc= UdmFindWordsSQL(A, CurRes, db);
+      if (rc == UDM_OK && CurRes->CoordList.ncoords)
+        rc= UdmUserCacheStoreSQL(A, CurRes, db);
+      break;
+#endif
+  }
+  if (!CurRes->PerSite)
+  {
+    size_t nbytes= sizeof(size_t) * CurRes->total_found;
+    CurRes->PerSite= nbytes ? (size_t*)UdmXmalloc(nbytes) : NULL;
+  }
+  return rc;
+}
+
+typedef struct search_param_st
+{
+  UDM_AGENT *Agent;
+  UDM_RESULT *Res;
+  UDM_DB *db;
+  int rc;
+  void *thd;
+} UDM_SPRM;
+
+
+static
+void *SearchThread(void *arg)
+{
+  UDM_SPRM *sprm= (UDM_SPRM*) arg;
+  sprm->rc= UdmFindWordsDB(sprm->Agent, sprm->Res, sprm->db);
+  return NULL;
+}
+
 int UdmFindWords(UDM_AGENT *A, UDM_RESULT *Res)
 {
-  const char      *cache_mode  = UdmVarListFindStr(&A->Conf->Vars, "Cache", "no");
-  size_t i, dbfrom = 0, dbto =  A->Conf->dbl.nitems;
-  int res = UDM_OK;
-  size_t    nwrd=0;
-  size_t    nwrdX[256];
-  size_t          *PerSite[256], *persite, *curpersite;
-  size_t          ResultsLimit = UdmVarListFindUnsigned(&A->Conf->Vars, "ResultsLimit", 0);
-  UDM_URL_CRD  *wrdX[256];
-  UDM_URL_CRD  *wrd = NULL, *curwrd = NULL;
-  UDM_URLDATA     *udtX[256];
-  UDM_URLDATA     *udt = NULL, *curudt = NULL;
+  const char  *cache_mode= UdmVarListFindStr(&A->Conf->Vars, "Cache", "no");
+  size_t      i, dbfrom= 0, dbto= A->Conf->dbl.nitems;
+  int res=    UDM_OK;
+  size_t      ResultsLimit = UdmVarListFindUnsigned(&A->Conf->Vars, "ResultsLimit", 0);
 
   if( strcasecmp(cache_mode, "yes") || UdmSearchCacheFind(A, Res))
   {
     /* If not found in search cache       */
     /* Let's do actual search now:        */
     /* Get groupped by url_id words array */
-  
-    for (i = dbfrom; i < dbto; i++)
+    UDM_RESULT TmpRes[256];
+    UDM_SPRM search_param[256];
+    int use_threads= A->Conf->ThreadCreate != NULL && dbfrom + 1 != dbto;
+
+    use_threads= 0; /* TODO: link against libmysqlclient_r */
+    
+    for (i= dbfrom; i < dbto; i++)
     {
+      UDM_RESULT *CurRes= &TmpRes[i];
       UDM_DB *db = &A->Conf->dbl.db[i];
-      const char *dbaddr= UdmVarListFindStr(&db->Vars,"DBAddr","<noaddr>");
-      
-      UdmLog(A, UDM_LOG_DEBUG, "UdmFind for %s", dbaddr);
-      Res->CoordList.Coords = NULL;
-      Res->CoordList.Data = NULL;
-      Res->CoordList.ncoords = 0;
-      Res->total_found = 0;
-      switch(db->DBDriver)
+      CurRes[0]= Res[0];    /* Copy WWList and bool items */
+
+      if (use_threads)
       {
-        case UDM_DB_SEARCHD:
-          res = UdmFindWordsSearchd(A, Res, db);
-          break;
-#ifdef HAVE_SQL
-        default:
-          res = UdmFindWordsSQL(A, Res, db);
-          UdmUserCacheStoreSQL(A, Res, db);
-          break;
-#endif
+        void *thd;
+        UdmLog(A, UDM_LOG_DEBUG, "Starting thread[%d]", i + 1);
+        search_param[i].Agent= UdmAgentInit(NULL, A->Conf, i + 1);
+        search_param[i].Res= CurRes;
+        search_param[i].db= db;
+        A->Conf->ThreadCreate(&thd, SearchThread, (void*) &search_param[i]);
+        search_param[i].thd= thd;
       }
-      wrdX[i] = Res->CoordList.Coords;
-      udtX[i] = Res->CoordList.Data;
-      nwrdX[i] = Res->total_found;
-      nwrd += Res->total_found;
-      if ((PerSite[i] = Res->PerSite) == NULL)
+      else
+        UdmFindWordsDB(A, CurRes, db);
+    }
+    
+    if (use_threads)
+    {
+      for (i= dbfrom; i < dbto; i++)
       {
-        PerSite[i] = (nwrdX[i]) ? (size_t*)UdmXmalloc(sizeof(size_t) * nwrdX[i]) : NULL;
+        void *thd= search_param[i].thd;
+        UdmLog(A, UDM_LOG_DEBUG, "Joining thread[%d]", i + 1);
+        A->Conf->ThreadJoin(thd);
       }
     }
-
-    if (nwrd > 0)
+    
+    if (use_threads)
     {
-      curwrd=wrd=(UDM_URL_CRD*)UdmMalloc(sizeof(*wrd)*nwrd);
-      curudt=udt=(UDM_URLDATA*)UdmMalloc(sizeof(*udt)*nwrd);
-      Res->PerSite = persite = curpersite = (size_t*)UdmMalloc(sizeof(size_t) * nwrd);
-      for (i = dbfrom; i < dbto; i++)
+      for (i= dbfrom; i < dbto; i++)
+        UdmAgentFree(search_param[i].Agent);
+    }
+    
+    for (Res->total_found= 0, i= dbfrom; i < dbto; i++)
+      Res->total_found+= TmpRes[i].total_found;
+
+    if (Res->total_found > 0)
+    {
+      size_t offs;
+
+      Res->CoordList.Coords= (UDM_URL_CRD*)UdmMalloc(sizeof(UDM_URL_CRD)
+                                                     * Res->total_found);
+      Res->CoordList.Data= (UDM_URLDATA*)UdmMalloc(sizeof(UDM_URLDATA)
+                                                   * Res->total_found);
+      Res->PerSite= (size_t*)UdmMalloc(sizeof(size_t) * Res->total_found);
+      for (offs= 0, i= dbfrom; i < dbto; offs+= TmpRes[i].total_found, i++)
       {
-        if(wrdX[i])
+        UDM_RESULT  *CurRes= &TmpRes[i];
+        UDM_URL_CRD *CurCoords= TmpRes[i].CoordList.Coords;
+        if(CurCoords)
         {
           size_t j;
           
           /* Set machine number */
-          for(j=0;j<nwrdX[i];j++)
+          for(j= 0; j < CurRes->total_found; j++)
           {
             /* 
               We use (256-i) to sort a document from the first database
               before the same document from the second database.
             */
-            wrdX[i][j].coord = (wrdX[i][j].coord << 8) + (255 - (i & 255));
+            CurCoords[j].coord = (CurCoords[j].coord << 8) + (255 - (i & 255));
           }
           
-          memcpy(curwrd,wrdX[i],sizeof(*curwrd)*nwrdX[i]);
-          curwrd+=nwrdX[i];
-          UDM_FREE(wrdX[i]);
-          memcpy(curpersite, PerSite[i], sizeof(size_t) * nwrdX[i]);
-          curpersite += nwrdX[i];
-          UDM_FREE(PerSite[i]);
-          if (udtX[i] != NULL)
+          memcpy(Res->CoordList.Coords + offs,CurCoords,
+                 sizeof(CurCoords[0])*CurRes->total_found);
+          memcpy(Res->PerSite + offs, CurRes->PerSite,
+                 sizeof(size_t) * CurRes->total_found);
+          if (CurRes->CoordList.Data != NULL)
           {
-            memcpy(curudt, udtX[i], sizeof(*curudt) * nwrdX[i]);
+            memcpy(Res->CoordList.Data + offs, CurRes->CoordList.Data,
+                   sizeof(Res->CoordList.Data[0]) * CurRes->total_found);
           }
           else
           {
-            bzero(curudt, sizeof(*curudt) * nwrdX[i]);
+            bzero(Res->CoordList.Data + offs,
+                  sizeof(Res->CoordList.Data[0]) * CurRes->total_found);
           }
-          curudt += nwrdX[i];
-          UDM_FREE(udtX[i]);
         }
+        UDM_FREE(CurRes->CoordList.Coords);
+        UDM_FREE(CurRes->CoordList.Data);
+        UDM_FREE(CurRes->PerSite);
       }
-/*    if (Res->offset) UdmSortSearchWordsByWeight(wrd,nwrd);*/
     }
 
-    Res->total_found = Res->CoordList.ncoords = nwrd;
-    Res->CoordList.Coords = wrd;
-    Res->CoordList.Data = udt;
-    Res->num_rows=Res->CoordList.ncoords;
+    Res->num_rows= Res->CoordList.ncoords= Res->total_found;
     
     if (dbto - dbfrom > 1)
     {
@@ -809,8 +825,10 @@ UDM_RESULT * __UDMCALL UdmFind(UDM_AGENT *A)
   size_t i, dbfrom = 0, dbto =  A->Conf->dbl.nitems;
   int    page_number = UdmVarListFindInt(&A->Conf->Vars, "np", 0);
   int    page_size   = UdmVarListFindInt(&A->Conf->Vars, "ps", 10);
+  size_t  ExcerptSize = (size_t)UdmVarListFindInt(&A->Conf->Vars, "ExcerptSize", 256);
+  size_t  ExcerptPadding = (size_t)UdmVarListFindInt(&A->Conf->Vars, "ExcerptPadding", 40);
   char    str[128];
-  
+
   UdmLog(A,UDM_LOG_DEBUG,"Start UdmFind");
   
   Res=UdmResultInit(NULL);
@@ -823,7 +841,7 @@ UDM_RESULT * __UDMCALL UdmFind(UDM_AGENT *A)
 
   if (!Res->total_found)
   {
-    int suggest= UdmVarListFindInt(&A->Conf->Vars, "Suggest", 0);
+    int suggest= UdmVarListFindBool(&A->Conf->Vars, "Suggest", 0);
     if (suggest)
     {
       if(UDM_OK != (res= UdmResAction(A, Res, UDM_RES_ACTION_SUGGEST)))
@@ -833,9 +851,16 @@ UDM_RESULT * __UDMCALL UdmFind(UDM_AGENT *A)
   
   UdmVarListReplaceStr(&A->Conf->Vars, "m", UdmVarListFindStr(&A->Conf->Vars, "orig_m", "all"));
   UdmVarListDel(&A->Conf->Vars, "orig_m");
+  UdmVarListReplaceInt(&A->Conf->Vars, "CurrentTimestamp", (int) time(0));
 
   Res->first = page_number * page_size;  
-  if(Res->first >= Res->total_found) Res->first = (Res->total_found)? (Res->total_found - 1) : 0;
+  if(Res->first >= Res->total_found)
+  {
+    /*Res->first = (Res->total_found)? (Res->total_found - 1) : 0;*/
+    Res->last= Res->first;
+    Res->num_rows= 0;
+    goto ret;
+  }
   if((Res->first+page_size)>Res->total_found)
   {
     Res->num_rows=Res->total_found-Res->first;
@@ -918,13 +943,27 @@ UDM_RESULT * __UDMCALL UdmFind(UDM_AGENT *A)
   UdmConvert(A->Conf, Res, A->Conf->lcs, A->Conf->bcs);
   ticks_ = UdmStartTimer() - ticks_;
   UdmLog(A, UDM_LOG_DEBUG, "Stop  UdmConvert:\t\t%.2f", (float)ticks_/1000);
-  
+
+  ticks_=UdmStartTimer();
+  UdmLog(A, UDM_LOG_DEBUG, "Start Excerpts");
+  for (i= 0; i < Res->num_rows; i++)
+  {
+    char *Excerpt;
+    if ((Excerpt= UdmExcerptDoc(A, Res, &Res->Doc[i], ExcerptSize, ExcerptPadding)))
+    {
+      UdmVarListReplaceStr(&Res->Doc[i].Sections, "body", Excerpt);
+      UdmFree(Excerpt);
+    }
+  }
+  ticks_ = UdmStartTimer() - ticks_;
+  UdmLog(A, UDM_LOG_DEBUG, "Stop  Excerpts:\t\t%.2f", (float)ticks_/1000);
+
+ret:  
+  UdmResWordInfo(A->Conf, Res);
   Res->work_time=ticks=UdmStartTimer()-ticks;
   UdmLog(A,UDM_LOG_DEBUG,"Done  UdmFind %.2f",(float)ticks/1000);
-  
   UdmTrack(A, Res);
 
-ret:
   if(res!=UDM_OK)
   {
     UdmResultFree(Res);
@@ -1215,7 +1254,7 @@ int UdmDBSetAddr(UDM_DB *db, const char *dbaddr, int mode)
       rc= UDM_ERROR;
       goto ret;
     }
-    UdmVarListReplaceStr(&db->Vars, "filename", s);
+    UdmVarListReplaceStr(&db->Vars, "filename", addr.filename);
   }
   else
   {
@@ -1354,6 +1393,8 @@ int UdmTrack(UDM_AGENT * query, UDM_RESULT *Res)
   int rc = UDM_OK;
 #ifdef HAVE_SQL
   size_t i, dbfrom = 0, dbto=  query->Conf->dbl.nitems; 
+  char * env= getenv("REMOTE_ADDR");
+  UdmVarListAddStr(&query->Conf->Vars, "IP", env ? env : "");
   
   for (i = dbfrom; i < dbto; i++)
   {
