@@ -40,6 +40,59 @@
 
 /*#define DEBUG_GUESSER*/
 
+typedef struct ct_match_st
+{
+  int offs_min; /* We'll use negative offsets for end-of-file offsets */
+  int offs_max; /* in the future.                                     */
+  const char *arg;
+  size_t length;
+  const char *content_type;
+} UDM_CTMATCH;
+
+static UDM_CTMATCH udm_magic[]=
+{
+  {0, 30, "<!DOCTYPE HTML", 14, "text/html"},
+  {0, 30, "<!doctype html", 14, "text/html"},
+  {0, 30, "<HEAD",          5,  "text/html"},
+  {0, 30, "<head",          5,  "text/html"},
+  {0, 30, "<TITLE",         6,  "text/html"},
+  {0, 30, "<title",         6,  "text/html"},
+  {0, 30, "<HTML",          5,  "text/html"},
+  {0, 30, "<html",          5,  "text/html"},
+  {0, 30, "<!--",           4,  "text/html"},
+  {0, 30, "<H1",            3,  "text/html"},
+  {0, 30, "<h1",            3,  "text/html"},
+  {0, 30, NULL, NULL}
+};
+
+const char *UdmGuessContentType(const char *begin, size_t length)
+{
+  UDM_CTMATCH *m;
+  const char *beg, *end, *cur;
+  for (m= udm_magic; m->content_type; m++)
+  {
+    beg= begin + m->offs_min;
+    end= begin + length - m->length;
+    if (end > begin + m->offs_max)
+      end= begin + m->offs_max;
+    for (cur= beg; cur < end; cur++)
+    {
+      if (!memcmp(cur, m->arg, m->length))
+        return m->content_type;
+    }
+  }
+  
+  /* Check if text/plain */
+  end= length > 128 ? begin + 128 : begin + length;
+  for (cur= begin; cur < end; cur++)
+  {
+    if (((unsigned char)cur[0]) < 9)
+      return "application/unknown";
+  }
+  return "text/plain";
+}
+
+
 static UDM_LANGMAP *FindLangMap(UDM_LANGMAPLIST *L, const char *lang, const char *charset, const char *filename) {
   UDM_LANGMAP *o = NULL;
   register size_t i;
