@@ -95,56 +95,115 @@
 
 #define MAXHSIZE	1023*4	/* TUNE */
 
+static size_t UdmDocPerSiteCached(UDM_AGENT *Indexer,
+                                  const char *hostinfo, size_t hostinfo_length)
+{
+  size_t i, res= 0;
+  
+  UDM_HREFLIST *Hrefs;
+  for (Hrefs= &Indexer->Conf->Hrefs, i= 0; i < Hrefs->nhrefs; i++)
+  {
+    UDM_HREF *Href= &Hrefs->Href[i];
+    if (Href->stored && Href->method != UDM_METHOD_DISALLOW &&
+        !strncmp(Href->url, hostinfo, hostinfo_length))
+      res++;
+  }
+  return res;
+}
+
+static int UdmCheckDocPerSite(UDM_AGENT *Indexer, UDM_HREF *Href, size_t *res,
+                              const char *hostinfo, size_t hostinfo_length)
+{
+  int rc= UDM_OK;
+  *res= UdmDocPerSiteCached(Indexer, hostinfo, hostinfo_length);
+
+  if (*res < Href->max_doc_per_site)
+  {
+    size_t doc_per_site_db;
+    UDM_DOCUMENT Doc;
+    UdmDocInit(&Doc);
+    UdmVarListAddStr(&Doc.Sections, "Hostinfo", hostinfo);
+    rc= UdmURLAction(Indexer, &Doc, UDM_URL_ACTION_DOCPERSITE);
+    doc_per_site_db= UdmVarListFindInt(&Doc.Sections, "DocPerSite", 0);
+    UdmDocFree(&Doc);
+    *res+= doc_per_site_db;
+  }
+  return rc;
+}
+
+
 __C_LINK int __UDMCALL UdmStoreHrefs(UDM_AGENT * Indexer)
 {
-  size_t    i, res;
+  size_t i;
+  int rc= UDM_OK;
   UDM_DOCUMENT  Doc;
-
+  UDM_HREFLIST *Hrefs;
+  size_t doc_per_site= 0;
+  size_t hostinfo_length= 0;
+  char hostinfo[128]= "";
+  
   UdmDocInit(&Doc);
 
-  for (i = 0; i < Indexer->Hrefs.dhrefs; i++)
+  UDM_GETLOCK(Indexer, UDM_LOCK_CONF);
+  
+  Hrefs= &Indexer->Conf->Hrefs;
+  
+  for (i = 0; i < Hrefs->nhrefs; i++)
   {
-    UDM_HREF *H = &Indexer->Hrefs.Href[i];
+    UDM_HREF *H = &Hrefs->Href[i];
+    int action= (i < Hrefs->dhrefs) ? UDM_URL_ACTION_ADD_LINK :
+                                      UDM_URL_ACTION_ADD;
     if(H->stored == 0)
     {
-      if( 1 /*strlen(H->url) <= UDM_URLSIZE */ )
-      {  /* FIXME: replace this by config parameter chacking */
-        UdmVarListReplaceInt(&Doc.Sections, "Referrer-ID", H->referrer);
-        UdmVarListReplaceUnsigned(&Doc.Sections,"Hops", H->hops);
-        UdmVarListReplaceStr(&Doc.Sections,"URL",H->url?H->url:"");
-        UdmVarListReplaceInt(&Doc.Sections, "URL_ID", UdmStrHash32(H->url ? H->url : ""));
-        UdmVarListReplaceInt(&Doc.Sections,"Site_id", H->site_id);
-        UdmVarListReplaceInt(&Doc.Sections,"Server_id", H->server_id);
-        if(UDM_OK != (res = UdmURLAction(Indexer, &Doc, UDM_URL_ACTION_ADD_LINK)))
+      if (H->max_doc_per_site)
+      {
+        if (!hostinfo[0] || strncmp(hostinfo, H->url, hostinfo_length))
         {
-          return(res);
+          /* New site */
+          UDM_URL url;
+          UdmURLInit(&url);
+          UdmURLParse(&url, H->url);
+          hostinfo_length= udm_snprintf(hostinfo, sizeof(hostinfo),
+                                       "%s://%s/", url.schema, url.hostinfo);
+          rc= UdmCheckDocPerSite(Indexer, H, &doc_per_site,
+                                 hostinfo, hostinfo_length);
+          UdmLog(Indexer, UDM_LOG_DEBUG, "DocPerSite: %d/%d",
+                 doc_per_site, H->max_doc_per_site);
+          UdmURLFree(&url);
+      
+          if (rc != UDM_OK)
+            goto ret;
+        }
+        else
+        {
+          /* Same site */
+          doc_per_site++;
+        }
+        if (doc_per_site > H->max_doc_per_site)
+        {
+          UdmLog(Indexer, UDM_LOG_DEBUG,
+                 "Too many docs (%d) per site, skip it", doc_per_site);
+          H->method = UDM_METHOD_DISALLOW;
+          H->stored= 1;
+          continue;
         }
       }
-      H->stored=1;
-    }
-  }
-  for (i = Indexer->Hrefs.dhrefs; i < Indexer->Hrefs.nhrefs; i++)
-  {
-    UDM_HREF *H = &Indexer->Hrefs.Href[i];
-    if(H->stored == 0)
-    {
-      if( 1 /*strlen(H->url) <= UDM_URLSIZE */ )
-      {  /* FIXME: replace this by config parameter chacking */
-        UdmVarListReplaceInt(&Doc.Sections, "Referrer-ID", H->referrer);
-        UdmVarListReplaceUnsigned(&Doc.Sections,"Hops", H->hops);
-        UdmVarListReplaceStr(&Doc.Sections,"URL",H->url?H->url:"");
-        UdmVarListReplaceInt(&Doc.Sections, "URL_ID", UdmStrHash32(H->url ? H->url : ""));
-        UdmVarListReplaceInt(&Doc.Sections, "HTDB_URL_ID", H->rec_id);
-        UdmVarListReplaceInt(&Doc.Sections,"Site_id", H->site_id);
-        UdmVarListReplaceInt(&Doc.Sections,"Server_id", H->server_id);
-        if(UDM_OK != (res = UdmURLAction(Indexer, &Doc, UDM_URL_ACTION_ADD)))
-        {
-          return(res);
-        }
-      }
-      H->stored=1;
+
+      UdmVarListReplaceInt(&Doc.Sections, "Referrer-ID", H->referrer);
+      UdmVarListReplaceUnsigned(&Doc.Sections,"Hops", H->hops);
+      UdmVarListReplaceStr(&Doc.Sections,"URL",H->url?H->url:"");
+      UdmVarListReplaceInt(&Doc.Sections, "URL_ID", UdmStrHash32(H->url ? H->url : ""));
+      UdmVarListReplaceInt(&Doc.Sections,"Site_id", H->site_id);
+      UdmVarListReplaceInt(&Doc.Sections,"Server_id", H->server_id);
+      UdmVarListReplaceInt(&Doc.Sections, "HTDB_URL_ID", H->rec_id);
+      if(UDM_OK != (rc= UdmURLAction(Indexer, &Doc, action)))
+        goto ret;
+      H->stored= 1;
     }
   }
+
+ret:
+
   UdmDocFree(&Doc);
   
   /*
@@ -152,7 +211,7 @@ __C_LINK int __UDMCALL UdmStoreHrefs(UDM_AGENT * Indexer)
     Note that it will became 0
     after next sort in AddUrl   
   */
-  Indexer->Hrefs.dhrefs = Indexer->Hrefs.nhrefs;
+  Hrefs->dhrefs = Hrefs->nhrefs;
   
   /*
     We should not free URL list with onw database
@@ -161,10 +220,12 @@ __C_LINK int __UDMCALL UdmStoreHrefs(UDM_AGENT * Indexer)
   */
   
   /* FIXME: this is incorrect with both SQL and built-in compiled */
-  if(Indexer->Hrefs.nhrefs > MAXHSIZE)
-    UdmHrefListFree(&Indexer->Hrefs);
+  if(Hrefs->nhrefs > MAXHSIZE)
+    UdmHrefListFree(&Indexer->Conf->Hrefs);
 
-  return UDM_OK;
+  UDM_RELEASELOCK(Indexer, UDM_LOCK_CONF);
+
+  return rc;
 }
 
 static void RelLink(UDM_URL *curURL, UDM_URL *newURL, char **str)
@@ -236,10 +297,13 @@ static int UdmDocBaseHref(UDM_AGENT *Indexer,UDM_DOCUMENT *Doc){
     
     UdmURLInit(&baseURL);
     
-    if(!(parse_res=UdmURLParse(&baseURL, basehref))) {
+    if(!(parse_res=UdmURLParse(&baseURL, basehref)) && baseURL.schema)
+    {
       UdmURLParse(&Doc->CurURL,basehref);
       UdmLog(Indexer, UDM_LOG_DEBUG, "BASE HREF '%s'", basehref);
-    }else{
+    }
+    else
+    {
       switch(parse_res){
       case UDM_URL_LONG:
         UdmLog(Indexer,UDM_LOG_ERROR,"BASE HREF too long: '%s'",basehref);
@@ -269,9 +333,7 @@ static int UdmFilterFind(UDM_MATCHLIST *L,const char *newhref,char *reason){
   }else{
     sprintf(reason,"Allow by default");
   }
-#ifdef USE_PARANOIA
   UdmViolationExit(paran);
-#endif
   return res;
 }
 
@@ -285,18 +347,23 @@ static int UdmSectionFilterFind(UDM_MATCHLIST *L, UDM_DOCUMENT *Doc, char *reaso
 #endif
 
   if((M=UdmMatchSectionListFind(L, Doc, NS, P))) {
-    sprintf(reason,"%s %s %s '%s'",M->arg,UdmMatchTypeStr(M->match_type),M->case_sense?"Sensitive":"InSensitive",M->pattern);
+    sprintf(reason,"%s %s %s %s '%s'", M->arg,
+            M->nomatch ? "NoMatch" : "Match",
+            UdmMatchTypeStr(M->match_type),
+            M->case_sense ? "Sensitive" : "InSensitive",
+            M->pattern);
     res=UdmMethod(M->arg);
   }else{
     sprintf(reason,"Allow by default");
   }
-#ifdef USE_PARANOIA
   UdmViolationExit(paran);
-#endif
   return res;
 }
 
-int UdmConvertHref(UDM_AGENT *Indexer, UDM_URL *CurURL, UDM_HREF *Href){
+
+int UdmConvertHref(UDM_AGENT *Indexer, UDM_URL *CurURL,
+                   UDM_SPIDERPARAM *Spider, UDM_HREF *Href)
+{
   int    parse_res, cascade;
   UDM_URL    newURL;
   char    *newhref = NULL;
@@ -306,6 +373,7 @@ int UdmConvertHref(UDM_AGENT *Indexer, UDM_URL *CurURL, UDM_HREF *Href){
   UDM_MATCH_PART  Parts[10];
   UDM_SERVER  *Srv;
   char    reason[1024]="";
+  int rc= UDM_OK;
   
   UdmURLInit(&newURL);
 
@@ -353,8 +421,22 @@ int UdmConvertHref(UDM_AGENT *Indexer, UDM_URL *CurURL, UDM_HREF *Href){
     Href->method=UDM_METHOD_DISALLOW;
     goto ret;
   }
+  {
+    int follow= UdmVarListFindInt(&Srv->Vars,"Follow",UDM_FOLLOW_PATH);
+    if (follow == UDM_FOLLOW_SITE && Spider->follow != UDM_FOLLOW_URLLIST &&
+        (strcasecmp(newURL.schema, CurURL->schema) ||
+         strcasecmp(newURL.hostinfo, CurURL->hostinfo)))
+    {
+      UdmLog(Indexer,UDM_LOG_DEBUG, "Skip: sites differ: %s %s",
+             newURL.hostinfo, CurURL->hostinfo);
+      Href->method= UDM_METHOD_DISALLOW;
+    }
+  }
   UDM_RELEASELOCK(Indexer, UDM_LOCK_CONF);
 
+  if (Href->method == UDM_METHOD_DISALLOW)
+    goto ret;
+  
   if (!strcasecmp(UDM_NULL2EMPTY(newURL.schema), "mailto") || !strcasecmp(UDM_NULL2EMPTY(newURL.schema), "javascript")) {
     UdmLog(Indexer,UDM_LOG_DEBUG,"'%s' schema, skip it", newURL.schema, newhref);
     Href->method=UDM_METHOD_DISALLOW;
@@ -378,6 +460,8 @@ int UdmConvertHref(UDM_AGENT *Indexer, UDM_URL *CurURL, UDM_HREF *Href){
     UdmLog(Indexer,UDM_LOG_DEBUG,"%s",reason);
   }
   
+  Href->max_doc_per_site= UdmVarListFindInt(&Srv->Vars, "MaxDocPerSite", 0);
+
   /* FIXME: add MaxHops, Robots */
   UDM_FREE(Href->url);
   Href->url = (char*)UdmStrdup(newhref);
@@ -386,24 +470,23 @@ ret:
   UDM_FREE(newhref);
   UDM_FREE(alias);
   UdmURLFree(&newURL);
-#ifdef USE_PARANOIA
   UdmViolationExit(paran);
-#endif
-  return UDM_OK;
+  return rc;
 }
 
-static int UdmDocConvertHrefs(UDM_AGENT *Indexer, UDM_DOCUMENT *Doc){
-  size_t    i;
-  int    hops=UdmVarListFindInt(&Doc->Sections,"Hops",0);
-  urlid_t    url_id = (urlid_t)UdmVarListFindInt(&Doc->Sections, "ID", 0);
-  uint4           maxhops = UdmVarListFindUnsigned(&Doc->Sections, "MaxHops", 255);
+static int UdmDocConvertHrefs(UDM_AGENT *Indexer, UDM_DOCUMENT *Doc)
+{
+  size_t  i;
+  int     hops=UdmVarListFindInt(&Doc->Sections,"Hops",0);
+  urlid_t url_id = (urlid_t)UdmVarListFindInt(&Doc->Sections, "ID", 0);
+  uint4   maxhops = UdmVarListFindUnsigned(&Doc->Sections, "MaxHops", 255);
 
   for(i=0;i<Doc->Hrefs.nhrefs;i++){
     UDM_HREF  *Href=&Doc->Hrefs.Href[i];
     Href->hops=hops+1;
-    UdmConvertHref(Indexer,&Doc->CurURL,Href);
+    UdmConvertHref(Indexer, &Doc->CurURL, &Doc->Spider, Href);
     Href->referrer=url_id;
-    if (maxhops > Href->hops) {
+    if (maxhops >= Href->hops) {
       Href->stored = 0;
     } else {
       Href->stored = 1;
@@ -413,20 +496,26 @@ static int UdmDocConvertHrefs(UDM_AGENT *Indexer, UDM_DOCUMENT *Doc){
   return UDM_OK;
 }
 
-int UdmDocStoreHrefs(UDM_AGENT *Indexer, UDM_DOCUMENT *Doc){
+
+int UdmDocStoreHrefs(UDM_AGENT *Indexer, UDM_DOCUMENT *Doc)
+{
   size_t i;
-  
+
   if(Doc->method==UDM_METHOD_HEAD)
     return UDM_OK;
-  
+
   UdmDocBaseHref(Indexer,Doc);
   UdmDocConvertHrefs(Indexer,Doc);
-  for(i=0;i<Doc->Hrefs.nhrefs;i++){
+
+  UDM_GETLOCK(Indexer, UDM_LOCK_CONF);  
+  for(i=0;i<Doc->Hrefs.nhrefs;i++)
+  {
     UDM_HREF  *Href=&Doc->Hrefs.Href[i];
-    if(Href->method!=UDM_METHOD_DISALLOW) {
-      UdmHrefListAdd(&Indexer->Hrefs, Href);
-    }
+    if(Href->method != UDM_METHOD_DISALLOW)
+      UdmHrefListAdd(&Indexer->Conf->Hrefs, Href);
   }
+  UDM_RELEASELOCK(Indexer, UDM_LOCK_CONF);
+
   return UDM_OK;
 }
 
@@ -476,9 +565,7 @@ __C_LINK int __UDMCALL UdmURLFile(UDM_AGENT *Indexer, const char *fname,int acti
       UdmVarListReplaceStr(&Indexer->Conf->Vars, "ul", str);
       result = UdmURLAction(Indexer, NULL, UDM_URL_ACTION_EXPIRE);
       if(result!=UDM_OK) { UdmURLFree(&myurl); 
-#ifdef USE_PARANOIA
       UdmViolationExit(paran);
-#endif
       return(result); }
       UdmVarListDel(&Indexer->Conf->Vars, "ul");
       break;
@@ -486,9 +573,7 @@ __C_LINK int __UDMCALL UdmURLFile(UDM_AGENT *Indexer, const char *fname,int acti
       UdmVarListReplaceStr(&Indexer->Conf->Vars, "ul", str);
       result=UdmClearDatabase(Indexer);
       if(result!=UDM_OK) { UdmURLFree(&myurl); 
-#ifdef USE_PARANOIA
       UdmViolationExit(paran);
-#endif
       return(UDM_ERROR); }
       UdmVarListDel(&Indexer->Conf->Vars, "ul");
       break;
@@ -496,7 +581,7 @@ __C_LINK int __UDMCALL UdmURLFile(UDM_AGENT *Indexer, const char *fname,int acti
       UdmHrefInit(&Href);
       Href.url=str;
       Href.method=UDM_METHOD_GET;
-      UdmHrefListAdd(&Indexer->Hrefs, &Href);
+      UdmHrefListAdd(&Indexer->Conf->Hrefs, &Href);
       break;
     case UDM_URL_FILE_PARSE:
       res=UdmURLParse(&myurl, str);
@@ -512,9 +597,7 @@ __C_LINK int __UDMCALL UdmURLFile(UDM_AGENT *Indexer, const char *fname,int acti
           UdmLog(Indexer,UDM_LOG_ERROR,"Error in URL: '%s'",str);
         }
         UdmURLFree(&myurl);
-#ifdef USE_PARANOIA
         UdmViolationExit(paran);
-#endif
         return(UDM_ERROR);
       }
       break;
@@ -524,9 +607,7 @@ __C_LINK int __UDMCALL UdmURLFile(UDM_AGENT *Indexer, const char *fname,int acti
   if(url_file!=stdin)
     fclose(url_file);
   UdmURLFree(&myurl);
-#ifdef USE_PARANOIA
   UdmViolationExit(paran);
-#endif
   return(UDM_OK);
 }
 
@@ -767,13 +848,16 @@ static int UdmDocProcessResponseHeaders(UDM_AGENT *Indexer,UDM_DOCUMENT *Doc){
     UdmVarListAddStr(&Doc->Sections,"Content-Type","application/octet-stream");
   }
 
-  if ((var = UdmVarListFind(&Doc->Sections, "Location"))) {
-          UDM_URL newURL;
+  if ((var = UdmVarListFind(&Doc->Sections, "Location")))
+  {
+    UDM_URL newURL;
           
-          UdmURLInit(&newURL);
-    switch(UdmURLParse(&newURL, var->val)) {
+    UdmURLInit(&newURL);
+    switch(UdmURLParse(&newURL, var->val))
+    {
         case UDM_URL_OK:
-          if (UDM_NULL2EMPTY(newURL.schema) != NULL) {
+          if (UDM_NULL2EMPTY(newURL.schema) != NULL)
+          {
             UDM_HREF Href;
             UdmHrefInit(&Href);
             Href.url=var->val;
@@ -862,30 +946,67 @@ ex:
 }
 #endif
 
+static size_t UdmDSTRParse(UDM_DSTR *d, const char *fmt, UDM_VARLIST *L)
+{
+  for ( ; fmt[0]; fmt++)
+  {
+    const char *end;
+    if (fmt[0] == '$' && fmt[1] == '{' && (end= strchr(fmt,'}')))
+    {
+      char name[128];
+      size_t namelen;
+      fmt+=2;
+      namelen= end - fmt;
+      if (namelen < sizeof(name) - 1)
+      {
+        const char *val;
+        memcpy(name, fmt, namelen);
+        name[namelen]= '\0';
+        if ((val= UdmVarListFindStr(L, name, NULL)))
+          UdmDSTRAppendSTR(d, val);
+      }
+      fmt= end;
+    }
+    else
+      UdmDSTRAppend(d, fmt, 1);
+  }
+  return d->size_data;
+}
+
+
 static int UdmParseSections (UDM_AGENT *Indexer, UDM_DOCUMENT *Doc)
 {
-  UDM_MATCH       *Alias;
-  char buf[1024];
-  UDM_MATCH_PART  Parts[10];
-  size_t nparts = 10;
-  UDM_VAR *Sec;
-  UDM_TEXTITEM Item;
   size_t i;
   UDM_DSTR tbuf;
 
   UdmDSTRInit(&tbuf, 1024);
   for (i = 0; i < Indexer->Conf->SectionMatch.nmatches; i++)
   {
+    UDM_VAR *Sec;
     UDM_HTMLTOK tag;
-    const char *htok, *last;
+    const char *htok, *last, *src;
+    char buf[1024];
+    UDM_MATCH_PART  Parts[10];
+    size_t nparts = 10;
+    UDM_TEXTITEM Item;
+    UDM_MATCH *Alias= &Indexer->Conf->SectionMatch.Match[i];
 
-    Alias = &Indexer->Conf->SectionMatch.Match[i];
-    if (UdmMatchExec(Alias, Doc->Buf.content, Doc->Buf.content, nparts, Parts)) continue;
+    if (Alias->arg1)
+    {
+      UdmDSTRReset(&tbuf);
+      UdmDSTRParse(&tbuf, Alias->arg1, &Doc->Sections);
+      src= tbuf.data;
+    }
+    else
+      src= Doc->Buf.content;
 
-    Sec = UdmVarListFind(&Doc->Sections, Alias->section);
-    if (! Sec) continue;
+    if (UdmMatchExec(Alias, src, src, nparts, Parts))
+      continue;
 
-    UdmMatchApply(buf, sizeof(buf) - 1, Doc->Buf.content, Alias->arg, Alias, nparts, Parts);
+    if (!(Sec= UdmVarListFind(&Doc->Sections, Alias->section)))
+      continue;
+
+    UdmMatchApply(buf, sizeof(buf) - 1, src, Alias->arg, Alias, nparts, Parts);
 
     Item.href=NULL;
     Item.section=Sec->section;
@@ -901,8 +1022,8 @@ static int UdmParseSections (UDM_AGENT *Indexer, UDM_DOCUMENT *Doc)
 */
 
     UdmHTMLTOKInit(&tag);
-    htok= UdmHTMLToken(buf, &last, &tag);
-    do
+    for (htok= UdmHTMLToken(buf, &last, &tag); htok; 
+         htok = UdmHTMLToken(NULL, &last, &tag))
     {
       if (tag.type == UDM_HTML_TXT && ! tag.script && ! tag.comment && ! tag.style)
       {
@@ -911,7 +1032,7 @@ static int UdmParseSections (UDM_AGENT *Indexer, UDM_DOCUMENT *Doc)
           Item.str = tbuf.data;
           UdmTextListAdd(&Doc->TextList, &Item);
       }
-    } while ((htok = UdmHTMLToken(NULL, &last, &tag)));
+    }
   }
 
   UdmDSTRFree(&tbuf);
@@ -922,13 +1043,14 @@ static int UdmDocParseContent(UDM_AGENT * Indexer, UDM_DOCUMENT * Doc){
   
 #ifdef USE_PARSER
   UDM_PARSER  *Parser;
-  const int       status = UdmVarListFindInt(&Doc->Sections, "Status", 0);
+  const int   status = UdmVarListFindInt(&Doc->Sections, "Status", 0);
 #endif
   const char  *real_content_type=NULL;
   const char  *url=UdmVarListFindStr(&Doc->Sections,"URL","");
   const char  *ct=UdmVarListFindStr(&Doc->Sections,"Content-Type","");
   const char  *ce=UdmVarListFindStr(&Doc->Sections,"Content-Encoding","");
   int    result=UDM_OK;
+  UDM_VAR *Sec;
   
   if(!strcmp(UDM_NULL2EMPTY(Doc->CurURL.filename), "robots.txt")) return UDM_OK;
   
@@ -987,6 +1109,20 @@ static int UdmDocParseContent(UDM_AGENT * Indexer, UDM_DOCUMENT * Doc){
   
   if(!real_content_type)real_content_type=ct;
   UdmVarListAddStr(&Doc->Sections,"Parser.Content-Type",real_content_type);
+
+  /* Create "Content" variable if requested */
+  if ((Sec= UdmVarListFind(&Doc->Sections, "HTTP.Content")))
+  {
+    size_t content_size=Doc->Buf.size - (Doc->Buf.content-Doc->Buf.buf);
+    if (content_size)
+    {
+      Sec->val= (char*) UdmRealloc(Sec->val, content_size + 1);
+      memcpy(Sec->val, Doc->Buf.content, content_size);
+      Sec->val[content_size]= '\0';
+      Sec->curlen= content_size;
+    }
+  }
+  
   
 #ifdef HAVE_ZLIB
   if (Doc->method == UDM_METHOD_GET &&
@@ -1140,51 +1276,45 @@ static int UdmDocAddServExtraHeaders(UDM_SERVER *Server,UDM_DOCUMENT *Doc){
 
 
 static int UdmNextTarget(UDM_AGENT * Indexer,UDM_DOCUMENT *Result){
-  int  result=UDM_NOTARGET;
-  int     u;
-  
-  UDM_GETLOCK(Indexer, UDM_LOCK_CONF);
-  u = (Indexer->Conf->url_number <= 0);
-  UDM_RELEASELOCK(Indexer, UDM_LOCK_CONF);
+  int  rc= UDM_NOTARGET;
 
-  if (u) {
-    return UDM_NOTARGET;
-  }
-  
-  /* Load targets into memory cache */
   UDM_GETLOCK(Indexer, UDM_LOCK_THREAD);
   UDM_GETLOCK(Indexer, UDM_LOCK_CONF);
-  u = (Indexer->Conf->Targets.cur_row >= Indexer->Conf->Targets.num_rows);
-  UDM_RELEASELOCK(Indexer, UDM_LOCK_CONF);
-  if (u) {
-    result = UdmTargets(Indexer);
-    if(result!=UDM_OK) {
-      UDM_RELEASELOCK(Indexer, UDM_LOCK_THREAD);
-      return result;
-    }
+
+  if (Indexer->Conf->url_number <= 0)
+    goto unlock; /* UDM_NOTARGET */
+
+  /* Load targets into memory cache */
+  if (Indexer->Conf->Targets.cur_row >= Indexer->Conf->Targets.num_rows)
+  {
+    if(UDM_OK != (rc= UdmTargets(Indexer)))
+      goto unlock;
   }
-  
+
   /* Choose next target */
-  UDM_GETLOCK(Indexer, UDM_LOCK_CONF);
-  if(Indexer->Conf->Targets.num_rows && (Indexer->Conf->Targets.cur_row < Indexer->Conf->Targets.num_rows ) ) {
+  if(Indexer->Conf->Targets.num_rows &&
+     Indexer->Conf->Targets.cur_row < Indexer->Conf->Targets.num_rows)
+  {
     UDM_DOCUMENT *Doc=&Indexer->Conf->Targets.Doc[Indexer->Conf->Targets.cur_row];
-    
+
     UdmVarListReplaceLst(&Result->Sections,&Doc->Sections,NULL,"*");
     UdmVarListReplaceLst(&Result->Sections,&Indexer->Conf->Sections,NULL,"*");
     UdmVarListReplaceLst(&Result->RequestHeaders, &Doc->RequestHeaders, NULL, "*");
-    
+
     Indexer->Conf->Targets.cur_row++;
     Indexer->Conf->url_number--;
-    UDM_RELEASELOCK(Indexer, UDM_LOCK_CONF);
-    UDM_RELEASELOCK(Indexer, UDM_LOCK_THREAD);
-    return UDM_OK;
+    rc= UDM_OK;
   }
-  
+  else
+    rc= UDM_NOTARGET;
+
+unlock:
   UDM_RELEASELOCK(Indexer, UDM_LOCK_CONF);
   UDM_RELEASELOCK(Indexer, UDM_LOCK_THREAD);
-  return UDM_NOTARGET;
+  return rc;
 }
 
+
 static int UdmVarListLog(UDM_AGENT *A,UDM_VARLIST *V,int l,const char *pre){
   size_t h;
   if (UdmNeedLog(l)) {
@@ -1196,9 +1326,11 @@ static int UdmVarListLog(UDM_AGENT *A,UDM_VARLIST *V,int l,const char *pre){
   return UDM_OK;
 }
 
+
 static int UdmVarList2Spider(UDM_SPIDERPARAM *S,UDM_VARLIST *V){
   S->period    =UdmVarListFindInt(V, "Period",    UDM_DEFAULT_REINDEX_TIME);
   S->maxhops    =UdmVarListFindInt(V, "MaxHops",  UDM_DEFAULT_MAX_HOPS);
+  S->doc_per_site= UdmVarListFindInt(V, "MaxDocPerSite",  0);
   S->follow    =UdmVarListFindInt(V, "Follow",    UDM_FOLLOW_PATH);
   S->max_net_errors  =UdmVarListFindInt(V, "MaxNetErrors",  UDM_MAXNETERRORS);
   S->net_error_delay_time  =UdmVarListFindInt(V, "NetErrorDelayTime",UDM_DEFAULT_NET_ERROR_DELAY_TIME);
@@ -1225,29 +1357,27 @@ __C_LINK int __UDMCALL UdmIndexNextURL(UDM_AGENT *Indexer){
   
   UDM_THREADINFO(Indexer,"Selecting","");
 
-  if(UDM_OK==(result=UdmStoreHrefs(Indexer))) {
-          if (Indexer->action != UDM_OK) {
-#ifdef USE_PARANOIA
+  if(UDM_OK==(result=UdmStoreHrefs(Indexer)))
+  {
+    if (Indexer->action != UDM_OK)
+    {
       UdmViolationExit(paran);
-#endif
       return Indexer->action;
     }
     result = UdmNextTarget(Indexer, &Doc);
   }
   
-  if(result==UDM_NOTARGET){
+  if(result==UDM_NOTARGET)
+  {
     UdmDocFree(&Doc);  /* To free Doc.connp->connp */
-#ifdef USE_PARANOIA
     UdmViolationExit(paran);
-#endif
     return result;
   }
   
-  if(result!=UDM_OK){
+  if(result!=UDM_OK)
+  {
     UdmDocFree(&Doc);
-#ifdef USE_PARANOIA
     UdmViolationExit(paran);
-#endif
     return result;
   }
   
@@ -1309,6 +1439,7 @@ __C_LINK int __UDMCALL UdmIndexNextURL(UDM_AGENT *Indexer){
       UdmVarListReplaceInt(&Doc.Sections, "Site_id", UdmServerGetSiteId(Indexer, Server, &Doc.CurURL));
       UdmVarListReplaceInt(&Doc.Sections, "Server_id", Server->site_id);
       UdmVarListReplaceInt(&Doc.Sections, "MaxHops", Doc.Spider.maxhops);
+      UdmVarListReplaceInt(&Doc.Sections, "MaxDocPerSite", Doc.Spider.doc_per_site);
       
       if(alstr != NULL) {
         /* Server Primary alias found */
@@ -1351,11 +1482,10 @@ __C_LINK int __UDMCALL UdmIndexNextURL(UDM_AGENT *Indexer){
 /*  UDM_RELEASELOCK(Indexer,UDM_LOCK_CONF);*/
   
   
-  if(result!=UDM_OK){
+  if(result!=UDM_OK)
+  {
     UdmDocFree(&Doc);
-#ifdef USE_PARANOIA
     UdmViolationExit(paran);
-#endif
     return result;
   }
   
@@ -1466,12 +1596,12 @@ __C_LINK int __UDMCALL UdmIndexNextURL(UDM_AGENT *Indexer){
   }
   
   
-  if(result!=UDM_OK){
-          UDM_FREE(origurl); UDM_FREE(aliasurl);
+  if(result!=UDM_OK)
+  {
+    UDM_FREE(origurl);
+    UDM_FREE(aliasurl);
     UdmDocFree(&Doc);
-#ifdef USE_PARANOIA
     UdmViolationExit(paran);
-#endif
     return result;
   }
 
@@ -1521,12 +1651,12 @@ __C_LINK int __UDMCALL UdmIndexNextURL(UDM_AGENT *Indexer){
         UdmURLParse(&Doc.CurURL,origurl);
       }
       
-      if(result!=UDM_OK){
-              UDM_FREE(origurl); UDM_FREE(aliasurl);
+      if(result!=UDM_OK)
+      {
+        UDM_FREE(origurl);
+        UDM_FREE(aliasurl);
         UdmDocFree(&Doc);
-#ifdef USE_PARANOIA
         UdmViolationExit(paran);
-#endif
         return result;
       }
       
@@ -1554,12 +1684,12 @@ __C_LINK int __UDMCALL UdmIndexNextURL(UDM_AGENT *Indexer){
     /* This is to give a chance for  */
     /* a concurent thread to take it */
     result=UdmDocStoreHrefs(Indexer,&Doc);
-    if(result!=UDM_OK){
-            UDM_FREE(origurl); UDM_FREE(aliasurl);
+    if(result!=UDM_OK)
+    {
+      UDM_FREE(origurl);
+      UDM_FREE(aliasurl);
       UdmDocFree(&Doc);
-#ifdef USE_PARANOIA
       UdmViolationExit(paran);
-#endif
       return result;
     }
     
@@ -1580,12 +1710,12 @@ __C_LINK int __UDMCALL UdmIndexNextURL(UDM_AGENT *Indexer){
       UDM_THREADINFO(Indexer,"Parsing",url);
       
       result = UdmDocParseContent(Indexer, &Doc);
-      if(result!=UDM_OK){
-              UDM_FREE(origurl); UDM_FREE(aliasurl);
+      if(result!=UDM_OK)
+      {
+        UDM_FREE(origurl);
+        UDM_FREE(aliasurl);
         UdmDocFree(&Doc);
-#ifdef USE_PARANOIA
         UdmViolationExit(paran);
-#endif
         return result;
       }
       
@@ -1680,10 +1810,9 @@ __C_LINK int __UDMCALL UdmIndexNextURL(UDM_AGENT *Indexer){
         }  
       }
       UDM_RELEASELOCK(Indexer,UDM_LOCK_CONF);
-      if(UDM_OK != (result = UdmURLAction(Indexer, &Doc, UDM_URL_ACTION_LINKS_DELETE))) {
-#ifdef USE_PARANOIA
+      if(UDM_OK != (result = UdmURLAction(Indexer, &Doc, UDM_URL_ACTION_LINKS_DELETE)))
+      {
         UdmViolationExit(paran);
-#endif
         return result;
       }
     }
@@ -1694,12 +1823,12 @@ __C_LINK int __UDMCALL UdmIndexNextURL(UDM_AGENT *Indexer){
       result=UDM_ERROR;
   }
   
-  if(result!=UDM_OK){
-          UDM_FREE(origurl); UDM_FREE(aliasurl);
+  if(result!=UDM_OK)
+  {
+    UDM_FREE(origurl);
+    UDM_FREE(aliasurl);
     UdmDocFree(&Doc);
-#ifdef USE_PARANOIA
     UdmViolationExit(paran);
-#endif
     return result;
   }
   
@@ -1716,8 +1845,6 @@ __C_LINK int __UDMCALL UdmIndexNextURL(UDM_AGENT *Indexer){
   UDM_GETLOCK(Indexer, UDM_LOCK_THREAD);
   if (result == UDM_OK) result = Indexer->action;
   UDM_RELEASELOCK(Indexer, UDM_LOCK_THREAD);
-#ifdef USE_PARANOIA
   UdmViolationExit(paran);
-#endif
   return result;
 }
