@@ -52,6 +52,9 @@
 #include "udm_uniconv.h"
 #include "udm_parsexml.h"
 
+#ifdef WIN32
+#include "udm_winutils.h"
+#endif
 /* This should be last include */
 #ifdef DMALLOC
 #include "dmalloc.h"
@@ -79,20 +82,30 @@ static UDM_ENV Conf;
 UDM_AGENT **ThreadIndexers;
 
 #ifdef HAVE_PTHREAD
-
 #ifdef WIN32
-
 HANDLE *threads;
-
-/* CallBack function for Thread information */
-extern __C_LINK void __UDMCALL UdmShowInfo(UDM_AGENT* A, const char *state, const char* str);
-
+#ifdef CONSOLE
+#define getpid(x) _getpid(x)
+#define localtime_r(x, y) localtime(x)
+#define strptime(x, y, z) (0)
+#endif
 #else
-
+#define CONSOLE
 pthread_t *threads;
-
+#endif
 #endif
 
+/* CallBack function for log information */
+#ifdef WIN32
+#ifdef CONSOLE
+#define UDM_DBTYPE "odbc"
+void UdmShowInfo(UDM_AGENT* A, const char *state, const char* str)
+{
+  printf("%d %s %s\n",A->handle,state,str);
+}
+#else
+extern __C_LINK void __UDMCALL UdmShowInfo(UDM_AGENT* A, const char *state, const char* str);
+#endif
 #endif
 
 
@@ -291,13 +304,14 @@ __C_LINK static int __UDMCALL ShowReferers(UDM_AGENT * Indexer)
   return(res);
 }
 
+
 #undef THINFO_TEST
 #ifdef THINFO_TEST
 /* CallBack function for Thread information */
-/*void UdmShowInfo(int handle,char *state, char* str)
+void UdmThreadInfo(int handle,char *state, char* str)
 {
-     printf("%d %s %s\n",handle,state,str);
-}*/
+  printf("%d %s %s\n",handle,state,str);
+}
 #endif
 
 
@@ -509,13 +523,13 @@ static void UdmFeatures(UDM_VARLIST *V)
 #endif
 }
 
-#ifndef WIN32
+#ifdef CONSOLE
 
 static int usage(int level)
 {
   fprintf(stderr, "\n");
   fprintf(stderr, "indexer from %s-%s-%s\n", PACKAGE, VERSION, UDM_DBTYPE);
-  fprintf(stderr, "http://www.mnogosearch.org/ (C)1998-2004, LavTech Corp.\n");
+  fprintf(stderr, "http://www.mnogosearch.org/ (C)1998-2005, LavTech Corp.\n");
   fprintf(stderr, "\n");
   fprintf(stderr, "Usage: indexer [OPTIONS]  [configfile]\n");
   fprintf(stderr, "\n");
@@ -603,7 +617,7 @@ static int usage(int level)
   return(0);
 }
 
-#endif /* ifndef WIN32 */
+#endif /* ifndef CONSOLE */
 
 UDM_AGENT Main;
 
@@ -727,7 +741,7 @@ static void UdmParseCmdLine(void)
           UdmHrefInit(&Href);
           Href.url=optarg;
           Href.method=UDM_METHOD_GET;
-          UdmHrefListAdd(&Main.Hrefs, &Href);
+          UdmHrefListAdd(&Conf.Hrefs, &Href);
         }
         break;
       case 'N': maxthreads=atoi(optarg);break;
@@ -801,8 +815,7 @@ static void * thread_main(void *arg)
         UdmEnvFree(&Conf);
         Conf=NewConf;
         UdmParseCmdLine();
-#ifdef WIN32
-#else
+#ifndef WIN32
         UdmOpenLog("indexer", &Conf, log2stderr);
 #endif
         UDM_RELEASELOCK(Indexer, UDM_LOCK_CONF);
@@ -997,7 +1010,7 @@ static void exitproc(void)
   unlink(pidname);
 }
 
-#ifndef WIN32
+#ifdef CONSOLE
 static char * time_pid_info(void)
 {
   struct tm * tim;
@@ -1082,9 +1095,6 @@ static int UdmClear(UDM_AGENT *A, const char *url_fname)
 
 static int UdmIndex(UDM_AGENT *A)
 {
-#ifdef WIN32
-  int int_res;  /* Temporary variable used to check return result */
-#endif
 
   ThreadIndexers = (UDM_AGENT**)UdmMalloc(maxthreads * sizeof(UDM_AGENT*));
                
@@ -1111,8 +1121,10 @@ static int UdmIndex(UDM_AGENT *A)
       UDM_RELEASELOCK(A, UDM_LOCK_CONF);
       UDM_RELEASELOCK(A, UDM_LOCK_THREAD);
 #ifdef WIN32
-      int_res = _beginthreadex(NULL, 0, &thread_main, NULL, 0, NULL);
-      assert(int_res != -1);
+      {
+        int int_res= _beginthreadex(NULL, 0, &thread_main, Indexer, 0, NULL);
+        assert(int_res != -1);
+      }
 #else
       {
         pthread_attr_t attr;
@@ -1285,7 +1297,7 @@ ex:
   return rc;
 }
 
-#ifndef WIN32
+#ifdef CONSOLE
 
 int main(int argc, char **argv)
 {
@@ -1312,7 +1324,7 @@ int main(int argc, char **argv)
   UdmSetLockProc(&Conf,UdmLockProc);
   UdmSetRefProc(&Conf,UdmRefProc);
 #ifdef THINFO_TEST
-  UdmSetThreadProc(&Conf,UdmShowInfo);
+  UdmSetThreadProc(&Conf,UdmShowThreadInfo);
 #endif
   UdmAgentInit(&Main,&Conf,0);
      
@@ -1353,13 +1365,22 @@ int main(int argc, char **argv)
     UdmEnvFree(&Conf);
     return(1);
   }
-     
+  
+#ifdef WIN32
+  {
+    char *conf_dir= WinUdmConfDir();
+    UdmVarListReplaceStr(&Conf.Vars,"ConfDir", conf_dir);
+    UdmVarListReplaceStr(&Conf.Vars,"ShareDir", conf_dir);
+    UDM_FREE(conf_dir);
+  }
+#else
   env= getenv("UDM_CONF_DIR");
   UdmVarListReplaceStr(&Conf.Vars,"ConfDir",env?env:UDM_CONF_DIR);
      
   env=getenv("UDM_SHARE_DIR");
   UdmVarListReplaceStr(&Conf.Vars,"ShareDir",env?env:UDM_SHARE_DIR);
-     
+#endif
+ 
   if (!*cname)
   {
     if (argc == 1)
@@ -1369,7 +1390,7 @@ int main(int argc, char **argv)
     }
     else
     {
-      const char *cd=UdmVarListFindStr(&Conf.Vars,"UDM_CONF_DIR",UDM_CONF_DIR);
+      const char *cd=UdmVarListFindStr(&Conf.Vars,"ConfDir",UDM_CONF_DIR);
       udm_snprintf(cname,sizeof(cname),"%s%s%s",cd,UDMSLASHSTR,"indexer.conf");
       cname[sizeof(cname)-1]='\0';
     }
@@ -1564,4 +1585,4 @@ ex:
   return(0);
 }
 
-#endif /* ifndef WIN32 */
+#endif /* ifdef CONSOLE */
