@@ -104,17 +104,13 @@ int UdmPrepareWords(UDM_AGENT * Indexer,UDM_DOCUMENT * Doc){
   
   if ((uword = (int*)UdmMalloc((uwordlen + 1) * sizeof(int))) == NULL)
   {
-#ifdef USE_PARANOIA
     UdmViolationExit(paran);
-#endif
     return UDM_ERROR;
   }
   if ((lcsword = (char*)UdmMalloc(12 * uwordlen + 1)) == NULL)
   {
     UDM_FREE(uword); 
-#ifdef USE_PARANOIA
     UdmViolationExit(paran);
-#endif
     return UDM_ERROR;
   }
 
@@ -164,9 +160,7 @@ int UdmPrepareWords(UDM_AGENT * Indexer,UDM_DOCUMENT * Doc){
     {
       UdmLog(Indexer, UDM_LOG_ERROR, "%s:%d Can't alloc %u bytes", __FILE__, __LINE__, dstlen);
       UDM_FREE(uword); UDM_FREE(lcsword);
-#ifdef USE_PARANOIA
       UdmViolationExit(paran);
-#endif
       return UDM_ERROR;
     }
     
@@ -178,9 +172,7 @@ int UdmPrepareWords(UDM_AGENT * Indexer,UDM_DOCUMENT * Doc){
     {
       UdmLog(Indexer, UDM_LOG_ERROR, "%s:%d Can't UdmUniDup", __FILE__, __LINE__);
       UDM_FREE(uword); UDM_FREE(lcsword); UDM_FREE(ustr);
-#ifdef USE_PARANOIA
       UdmViolationExit(paran);
-#endif
       return UDM_ERROR;
     }
     
@@ -225,17 +217,13 @@ int UdmPrepareWords(UDM_AGENT * Indexer,UDM_DOCUMENT * Doc){
             if ((uword = (int*)UdmRealloc(uword, (uwordlen + 1) * sizeof(int))) == NULL)
             { 
               UDM_FREE(lcsword); UDM_FREE(ustr); UDM_FREE(UStr); 
-#ifdef USE_PARANOIA
               UdmViolationExit(paran);
-#endif
               return UDM_ERROR;
             }
             if ((lcsword = (char*)UdmRealloc(lcsword, 12 *uwordlen + 1)) == NULL)
             { 
               UDM_FREE(uword); UDM_FREE(ustr); UDM_FREE(UStr); 
-#ifdef USE_PARANOIA
               UdmViolationExit(paran);
-#endif
               return UDM_ERROR;
             }
           }
@@ -311,9 +299,7 @@ int UdmPrepareWords(UDM_AGENT * Indexer,UDM_DOCUMENT * Doc){
   UdmVarListReplaceInt(&Doc->Sections,"crc32",crc32);
   
   UDM_FREE(uword); UDM_FREE(lcsword);
-#ifdef USE_PARANOIA
   UdmViolationExit(paran);
-#endif
   return res;
 }
 
@@ -435,26 +421,51 @@ int UdmHTMLTOKInit(UDM_HTMLTOK *tag)
   return UDM_OK;
 }
 
+static char spacemap[256]=
+{
+  0,0,0,0,0,0,0,0,0,1,1,1,1,1,0,0,
+  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
+  1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
+  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
+  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
+  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
+  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
+  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
+  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
+  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
+  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
+  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
+  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
+  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
+  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
+  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0  
+};
+
+static inline int IsSpaceChar(int x)
+{
+  return (int) spacemap[x];
+}
+
 
 const char * UdmHTMLToken(const char * s, const char ** lt,UDM_HTMLTOK *t)
 {
   t->ntoks= 0;
   t->s= s;
-  t->lt= lt;
   
   if(t->s == NULL && (t->s = *lt) == NULL)
     return NULL;
 
-  if(!*t->s)
-    return NULL;
-  
-  if(!strncmp(t->s,"<!--",4))
-    t->type=UDM_HTML_COM;
-  else if(*t->s=='<')
-    t->type=UDM_HTML_TAG;
+  if (*t->s == '<')
+  {
+    t->type= strncmp(t->s + 1, "!--", 3) ? UDM_HTML_TAG : UDM_HTML_COM;
+  }
   else
+  {
+    if(!*t->s)
+      return NULL;
     t->type=UDM_HTML_TXT;
-
+  }
+  
   switch(t->type)
   {
     case UDM_HTML_TAG:
@@ -466,7 +477,7 @@ const char * UdmHTMLToken(const char * s, const char ** lt,UDM_HTMLTOK *t)
         size_t nt=t->ntoks;
         
         /* Skip leading spaces */
-        while((*t->b)&&strchr(" \t\r\n",*t->b)) (*t->next_b)(t);
+        for ( ; IsSpaceChar((unsigned char)*t->b) ; t->b++);
 
         if(*t->b=='>')
         {
@@ -481,7 +492,7 @@ const char * UdmHTMLToken(const char * s, const char ** lt,UDM_HTMLTOK *t)
         }
 
         /* Skip non-spaces, i.e. name */
-        for(t->e = t->b; (*t->e) && !strchr(" =>\t\r\n", *t->e); (*t->next_e)(t));
+        for(t->e = t->b; (*t->e) && !strchr(" =>\t\r\n", *t->e); t->e++);
         
         if(t->ntoks<UDM_MAXTAGVAL)
           t->ntoks++;
@@ -493,14 +504,26 @@ const char * UdmHTMLToken(const char * s, const char ** lt,UDM_HTMLTOK *t)
 
         if (nt == 0)
         {
-          if(!strncasecmp(t->b,"script",6)) t->script = 1;
-          if(!strncasecmp(t->b,"/script",7)) t->script = 0;
-          if(!strncasecmp(t->b, "noindex", 7)) t->comment = 1;
-          if(!strncasecmp(t->b, "/noindex", 8)) t->comment = 0;
-          if(!strncasecmp(t->b, "style", 5)) t->style = 1;
-          if(!strncasecmp(t->b, "/style", 6)) t->style = 0;
-          if(!strncasecmp(t->b, "body", 4)) t->body = 1;
-          if(!strncasecmp(t->b, "/body", 5)) t->body = 0;
+          const char *beg;
+          int val;
+          if (t->b[0] == '/')
+          {
+            val= 0;
+            beg= t->b + 1;
+          }
+          else
+          {
+            val= 1;
+            beg= t->b;
+          }
+          if (!strncasecmp(beg,"script",6))
+            t->script= val;
+          else if (!strncasecmp(beg, "noindex", 7))
+            t->comment= val;
+          else if (!strncasecmp(beg, "style", 5))
+            t->style= val;
+          else if (!strncasecmp(beg, "body", 4))
+            t->body= val;
         }
 
         if(*t->e=='>')
@@ -516,40 +539,40 @@ const char * UdmHTMLToken(const char * s, const char ** lt,UDM_HTMLTOK *t)
         }
         
         /* Skip spaces */
-        while((*t->e) && strchr(" \t\r\n",*t->e))(*t->next_e)(t);
+        for( ; IsSpaceChar((unsigned char) *t->e) ; t->e++);
         
         if(*t->e != '=')
         {
           t->b = t->e;
-               *lt = t->b;        /* bug when hang on broken inside tag pages fix */
+          *lt = t->b;        /* bug when hang on broken inside tag pages fix */
           continue;
         }
         
         /* Skip spaces */
-        for(t->b = t->e + 1; (*t->b) && strchr(" \r\n\t", *t->b); (*t->next_b)(t));
+        for(t->b = t->e + 1; IsSpaceChar((unsigned char) *t->b); t->b++);
         
         if(*t->b == '"')
         {
           t->b++;
           valbeg = t->b;
-          for(t->e = t->b; (*t->e) && (*t->e != '"'); (*t->next_e)(t));
+          for(t->e = t->b; (*t->e) && (*t->e != '"'); t->e++);
           valend = t->e;
           t->b = t->e;
-          if(*t->b == '"')(*t->next_b)(t);
+          if(*t->b == '"')t->b++;
         }
         else if(*t->b == '\'')
         {
           t->b++;
           valbeg = t->b;
-          for(t->e = t->b; (*t->e) && (*t->e != '\''); (*t->next_e)(t));
+          for(t->e = t->b; (*t->e) && (*t->e != '\''); t->e++);
           valend = t->e;
           t->b = t->e;
-          if(*t->b == '\'')(*t->next_b)(t);
+          if(*t->b == '\'')t->b++;
         }
         else
         {
           valbeg = t->b;
-          for(t->e = t->b; (*t->e) && !strchr(" >\t\r\n", *t->e);(*t->next_e)(t));
+          for(t->e = t->b; (*t->e) && !strchr(" >\t\r\n", *t->e);t->e++);
           valend = t->e;
           t->b = t->e;
         }
@@ -562,13 +585,15 @@ const char * UdmHTMLToken(const char * s, const char ** lt,UDM_HTMLTOK *t)
     case UDM_HTML_COM: /* comment */
       
       if(!strncasecmp(t->s, "<!--UdmComment-->",17) ||
-         !strncasecmp(t->s, "<!--noindex-->", 14))
+         !strncasecmp(t->s, "<!--noindex-->", 14) ||
+         !strncasecmp(t->s, "<!--X-BotPNI-->",15))
         t->comment=1;
       else if(!strncasecmp(t->s, "<!--/UdmComment-->",18) ||
-              !strncasecmp(t->s, "<!--/noindex-->", 15))
+              !strncasecmp(t->s, "<!--/noindex-->", 15) ||
+              !strncasecmp(t->s, "<!--X-BotPNI-End-->", 19))
         t->comment=0;
 
-      for(t->e = t->s; (*t->e) && (strncmp(t->e, "-->", 3)); (*t->next_e)(t));
+      for(t->e = t->s; (*t->e) && (strncmp(t->e, "-->", 3)); t->e++);
       if(!strncmp(t->e, "-->", 3)) *lt = t->e + 3;
       else  *lt = t->e;
       break;
@@ -580,7 +605,7 @@ const char * UdmHTMLToken(const char * s, const char ** lt,UDM_HTMLTOK *t)
       /* <script> x="<"; </script> */
       /* We should find </script>  */
       
-      for(t->e = t->s; *t->e; (*t->next_e)(t))
+      for(t->e = t->s; *t->e; t->e++)
       {
         if(*t->e == '<')
         {
@@ -635,23 +660,17 @@ int UdmHTMLParseTag(UDM_HTMLTOK * tag,UDM_DOCUMENT * Doc)
 
   if(!tag->ntoks)
   {
-#ifdef USE_PARANOIA
     UdmViolationExit(paran);
-#endif
     return(0);
   }
   if(!tag->toks[0].name)
   {
-#ifdef USE_PARANOIA
     UdmViolationExit(paran);
-#endif
     return(0);
   }
   if(tag->toks[0].nlen>sizeof(name)-1)
   {
-#ifdef USE_PARANOIA
     UdmViolationExit(paran);
-#endif
     return(0);
   }
 
@@ -888,9 +907,7 @@ int UdmHTMLParseTag(UDM_HTMLTOK * tag,UDM_DOCUMENT * Doc)
   UDM_FREE(lang);
   UDM_FREE(secname);
   
-#ifdef USE_PARANOIA
   UdmViolationExit(paran);
-#endif
   return 0;
 }
 
@@ -912,9 +929,14 @@ int UdmHTMLParse(UDM_AGENT *Indexer,UDM_DOCUMENT *Doc){
   UDM_VAR      *TSec=UdmVarListFind(&Doc->Sections,"title");
   int          body_sec  = BSec ? BSec->section : 0;
   int          title_sec = TSec ? TSec->section : 0;
+  int          body_sec_flags= BSec ? BSec->flags : 0;
+  int          title_sec_flags= TSec ? TSec->flags : 0;
   char         scb[]="body";
   char         sct[]="title";
-
+  /*
+    We cannot use BSec and TSec later, as they are
+    modified by UdmVarListReplace from UdmHTMLParseTag
+  */
 #ifdef USE_PARANOIA
   void *paran = UdmViolationEnter();
 #endif
@@ -944,8 +966,8 @@ int UdmHTMLParse(UDM_AGENT *Indexer,UDM_DOCUMENT *Doc){
         
         tmp = UdmStrndup(tmpbeg,(size_t)(tmpend-tmpbeg+1));
         
-        if (BSec && !(BSec->flags & UDM_VARFLAG_USERDEF) && !tag.comment &&
-            tag.body && !tag.script && !tag.style && tag.index)
+        if (body_sec && !(body_sec_flags & UDM_VARFLAG_USERDEF) &&
+            !tag.comment && tag.body && !tag.script && !tag.style && tag.index)
         {
           Item.href=tag.lasthref;
           Item.str=tmp;
@@ -958,8 +980,8 @@ int UdmHTMLParse(UDM_AGENT *Indexer,UDM_DOCUMENT *Doc){
 #endif
             UdmTextListAdd(&Doc->TextList,&Item);
         }
-        if (TSec && !(TSec->flags & UDM_VARFLAG_USERDEF) && !tag.comment &&
-            tag.title && tag.index)
+        if (title_sec && !(title_sec_flags & UDM_VARFLAG_USERDEF) &&
+            !tag.comment && tag.title && tag.index)
         {
           Item.href=NULL;
           Item.str=tmp;
@@ -978,8 +1000,6 @@ int UdmHTMLParse(UDM_AGENT *Indexer,UDM_DOCUMENT *Doc){
     
   }
   UDM_FREE(tag.lasthref);  
-#ifdef USE_PARANOIA
   UdmViolationExit(paran);
-#endif
   return UDM_OK;
 }
