@@ -326,13 +326,13 @@ static int UdmHTTPSGet(UDM_AGENT *Indexer,UDM_DOCUMENT *Doc)
 
     SSL_set_fd (ssl, fd);
 
-    if (SSL_connect(ssl)==-1){
+    if (SSL_connect(ssl) < 0){
       sslcleanup;
       return UDM_NET_ERROR;
     }
 
     /* Send HTTP request */
-    if ((SSL_write(ssl, Doc->Buf.buf, (int)strlen(Doc->Buf.buf))) == -1){
+    if ((SSL_write(ssl, Doc->Buf.buf, (int)strlen(Doc->Buf.buf))) < 0){
       sslcleanup;
       return UDM_NET_ERROR;
     }
