@@ -97,109 +97,6 @@ static char * BuildPageURL(UDM_VARLIST * vars, char **dst)
   return NULL;
 }
 
-/*****************************************************************/
-
-static int WordInfo(UDM_ENV *Env, UDM_RESULT *Res)
-{
-  size_t  len, i, j, wsize;
-  char  *wordinfo= NULL, *end;
-  size_t  corder= (size_t)-1, ccount= 0;
-  int have_suggestions= 0;
-  
-  for(len = i = 0; i < Res->WWList.nwords; i++) 
-    len += Res->WWList.Word[i].len + 64;
-  
-  wsize= (1+len)*sizeof(char);
-  wordinfo= (char*) UdmMalloc(wsize);
-  *wordinfo= '\0';
-  
-  for(i = 0; i < Res->WWList.nwords; i++)
-  {
-    if (Res->WWList.Word[i].count > 0 || 
-        Res->WWList.Word[i].origin == UDM_WORD_ORIGIN_QUERY)
-    {
-      if(wordinfo[0])
-        strcat(wordinfo,", ");
-      sprintf(UDM_STREND(wordinfo)," %s : %d", Res->WWList.Word[i].word, Res->WWList.Word[i].count);
-    }
-    else if (Res->WWList.Word[i].origin == UDM_WORD_ORIGIN_STOP)
-    {
-      if(wordinfo[0])
-        strcat(wordinfo,", ");
-      sprintf(UDM_STREND(wordinfo)," %s : stopword", Res->WWList.Word[i].word);
-    }
-  }
-  
-  UdmVarListAddStr(&Env->Vars, "WE", wordinfo);
-  
-  *wordinfo = '\0';
-  for(i = 0; i < Res->WWList.nwords; i++)
-  {
-    corder= Res->WWList.Word[i].order;
-    ccount= 0;
-    for(j= 0; j < Res->WWList.nwords; j++)
-      if (Res->WWList.Word[j].order == corder)
-        ccount += Res->WWList.Word[j].count;
-    if (Res->WWList.Word[i].origin == UDM_WORD_ORIGIN_STOP)
-    {
-      sprintf(UDM_STREND(wordinfo),"%s%s : stopword", (*wordinfo) ? ", " : "",  Res->WWList.Word[i].word);
-    }
-    else if (Res->WWList.Word[i].origin == UDM_WORD_ORIGIN_QUERY)
-    {
-      sprintf(UDM_STREND(wordinfo),"%s%s : %d / %d", 
-        (*wordinfo) ? ", " : "", Res->WWList.Word[i].word, Res->WWList.Word[i].count, ccount);
-    }
-  }
-  UdmVarListAddStr(&Env->Vars, "W", wordinfo);
-  
-  *wordinfo= '\0';
-  end= wordinfo;
-  for (i= 0; i < Res->WWList.nwords; i++)
-  {
-    UDM_WIDEWORD *Wi= &Res->WWList.Word[i];
-    UDM_WIDEWORD *Wb= NULL;
-   
-    if (Wi->origin == UDM_WORD_ORIGIN_QUERY)
-    {
-      if (Wi->count > 0)
-      {
-        Wb= Wi;
-      }
-      else
-      {
-        ccount= 0;
-        for (j= 0; j < Res->WWList.nwords; j++)
-        {
-          UDM_WIDEWORD *Wj= &Res->WWList.Word[j];
-          if (Wj->origin == UDM_WORD_ORIGIN_SUGGEST &&
-              Wj->order == Wi->order && Wj->count > ccount)
-          {
-            ccount= Res->WWList.Word[j].count;
-            Wb= Wj;
-            have_suggestions= 1;
-          }
-        }
-      }
-    }
-    else if (Wi->origin == UDM_WORD_ORIGIN_STOP)
-    {
-      Wb= Wi;
-    }
-    
-    if (Wb)
-    {
-      sprintf(end, "%s%s", wordinfo[0] ? " " : "", Wb->word);
-      end= end + strlen(end);
-    }
-  }
-  
-  if (have_suggestions)
-    UdmVarListAddStr(&Env->Vars, "WS", wordinfo);
-  UDM_FREE(wordinfo);
-  return UDM_OK;
-}
-  
-
 /*****************************************************************/
 
 static char *UdmConfDir(void)
@@ -369,6 +266,20 @@ static void UdmWSACleanup(void)
 #endif
 }
 /************************************************************/
+#if defined(HAVE_PTHREAD) && !defined(WIN32)
+static
+int UdmThreadCreate(void **thd, void *(*start_routine)(void *), void *arg)
+{
+  return pthread_create((pthread_t*) thd, NULL, start_routine, arg); 
+}
+
+static
+int UdmThreadJoin(void *thd)
+{
+  return pthread_join((pthread_t) thd, NULL);
+}        
+#endif
+
 int main(int argc, char ** argv)
 {
   const char  *env, *bcharset, *lcharset;
@@ -378,7 +289,7 @@ int main(int argc, char ** argv)
   char        *query_string= NULL, *nav= NULL, *url= NULL;
   char        *searchwords= NULL, *storedstr= NULL;
   int         page_size, page_number, res,httpd= 0;
-  size_t      i, swlen= 0, nav_len, ExcerptSize, ExcerptPadding, storedlen, catcolumns= 0;
+  size_t      i, swlen= 0, nav_len, storedlen, catcolumns= 0;
   ssize_t     page1,page2,npages,ppp=10;
   UDM_ENV     *Env;
   UDM_AGENT   *Agent;
@@ -404,6 +315,10 @@ int main(int argc, char ** argv)
   
   UdmInit();
   Env=UdmEnvInit(NULL);
+#if defined(HAVE_PTHREAD) && !defined(WIN32)
+  Env->ThreadCreate= UdmThreadCreate;
+  Env->ThreadJoin= UdmThreadJoin;
+#endif
   UdmVarListInit(&tmpl);
   UdmVarListInit(&query_vars);
   Agent = UdmAgentInit(NULL, Env, 0);
@@ -535,6 +450,21 @@ int main(int argc, char ** argv)
     UdmAgentFree(Agent);
     return(0);
   }
+
+#ifdef HAVE_SETVBUF
+  {
+    /*
+      Set stdout buffer size. It allows to avoid re-rendering
+      in the browser. After setting StdoutBufferSize to a
+      reasonably big value, search.cgi will send whole output
+      at once in the end, rather than gradually.
+    */
+    size_t bsz= (size_t) UdmVarListFindInt(&Env->Vars, "StdoutBufferSize", 0);
+    if (bsz > 0)
+      setvbuf(stdout, NULL, _IOFBF, bsz);
+  }
+#endif
+
   
   UdmTemplatePrint(Agent, NULL, NULL, 0, &Env->Vars, &tmpl, "variables");
 
@@ -548,8 +478,6 @@ int main(int argc, char ** argv)
   UdmVarListReplaceStr(&Env->Vars, "tmplt", template_name);
   UdmVarListAddStr(&Env->Vars,"QUERY_STRING",query_string);
   UdmVarListAddStr(&Env->Vars,"self",self);
-  env= getenv("REMOTE_ADDR");
-  UdmVarListAddStr(&Env->Vars, "IP", env?env:"");
   
   bcharset= UdmVarListFindStr(&Env->Vars,"BrowserCharset","iso-8859-1");
   Env->bcs= UdmGetCharSet(bcharset);
@@ -676,13 +604,11 @@ int main(int argc, char ** argv)
       goto freeres;
     goto end;
   }
-  
   UdmVarListAddInt(&Env->Vars,"first",(int)Res->first);
   UdmVarListAddInt(&Env->Vars,"last",(int)Res->last);
   UdmVarListAddInt(&Env->Vars,"total",(int)Res->total_found);
   sprintf(search_time,"%.3f",((double)Res->work_time)/1000);
   UdmVarListAddStr(&Env->Vars,"SearchTime",search_time);
-  WordInfo(Env,Res);
   UdmTemplatePrint(Agent, stdout, NULL, 0, &Env->Vars, &tmpl, "top");
 #ifdef TRIAL_VER
   fprintf(stdout,"%s",POWERED_LOGO);
@@ -792,15 +718,13 @@ int main(int argc, char ** argv)
   
   UdmTemplatePrint(Agent, stdout, NULL, 0, &Env->Vars,&tmpl, "restop");
   
-  ExcerptSize = (size_t)UdmVarListFindInt(&Env->Vars, "ExcerptSize", 256);
-  ExcerptPadding = (size_t)UdmVarListFindInt(&Env->Vars, "ExcerptPadding", 40);
-  
   for(i=0;i<Res->num_rows;i++)
   {
     UDM_DOCUMENT *Doc=&Res->Doc[i];
     UDM_CATEGORY C;
     UDM_MATCH    *Alias;
-    char         *aliastr, *alcopy, *clist, *eu, *Excerpt;
+    UDM_MATCH_PART Parts[10];
+    char         *aliastr, *alcopy, *clist, *eu;
     const char   *u;
     size_t       cl, sc, clistsize;
     urlid_t      dc_url_id = (urlid_t)UdmVarListFindInt(&Doc->Sections, "ID", 0);
@@ -808,10 +732,11 @@ int main(int argc, char ** argv)
     
     /* Create "Alias" variable */
     alcopy=UdmRemoveHiLightDup(UdmVarListFindStr(&Doc->Sections,"URL",""));
-    if ((Alias=UdmMatchListFind(&Agent->Conf->Aliases,alcopy,0,NULL)))
+    if ((Alias=UdmMatchListFind(&Agent->Conf->Aliases,alcopy,10,Parts)))
     {
-      aliastr= (char*)UdmMalloc(strlen(Alias->arg) + strlen(alcopy) + 1);
-      sprintf(aliastr, "%s%s", Alias->arg, alcopy + strlen(Alias->pattern));
+      size_t nbytes= strlen(Alias->arg) + strlen(alcopy) + 64;
+      aliastr= (char*)UdmMalloc(nbytes);
+      UdmMatchApply(aliastr, nbytes, alcopy, Alias->arg, Alias, 10, Parts);
     }
     else
     {
@@ -902,10 +827,6 @@ int main(int argc, char ** argv)
     UdmFree(eu);
     
     UdmVarListReplaceStr(&Env->Vars, "stored_href", UdmVarListFindStr(&Doc->Sections, "CachedCopy", NULL) ? storedstr : "");
-    if ((Excerpt= UdmExcerptDoc(Agent, Res, Doc, ExcerptSize, ExcerptPadding)))
-      UdmVarListReplaceStr(&Env->Vars, "body", Excerpt);
-    UDM_FREE(Excerpt);
-    
     if ( (sc = UdmVarListFindInt(&Env->Vars, "site", 0)) == 0)
     {
       UdmVarListReplaceOrDelInt(&query_vars,"np", 0);
