@@ -712,8 +712,8 @@ int UdmPrepare(UDM_AGENT * query,UDM_RESULT *Res)
   ns is between 1..256
 */
 static
-inline uint4 UdmCalcCosineWeight(unsigned int *R, unsigned int *D,
-                                 size_t num, float Rsum_factor)
+inline uint4 UdmCalcCosineWeight(unsigned int *R, unsigned int *D, size_t num,
+                                 float Rsum_factor, float nwords_factor)
 {
   size_t i;
   float res;
@@ -730,7 +730,7 @@ inline uint4 UdmCalcCosineWeight(unsigned int *R, unsigned int *D,
     }
   }
   
-  res= Rsum_factor * (float) RDsum / sqrt(Dsum) + 0.5;
+  res= Rsum_factor * nwords_factor * (float) RDsum / sqrt(Dsum) + 0.5;
   
 #if 0
   fprintf(stderr,
@@ -751,20 +751,50 @@ static int UdmOriginWeight(int origin)
   return 0;
 }
 
+
 static unsigned int CalcAverageWordDistance(UDM_PHR_DAT *phr, size_t num)
 {
-  size_t i, sum = 0;
-  size_t np = 0, cur_order;
+  size_t i, sum= 0, np= 0;
 
-  UdmSort((void*)phr, num, sizeof(UDM_PHR_DAT), (qsort_cmp)cmpphr);
-  cur_order = phr[0].order;
-  for (i = 1; i < num; i++)
+  if (num < 2)
+    return 0;
+  
+  if (num == 2)
   {
-    if (phr[i].order != cur_order)
+    return phr[0].order == phr[1].order ? 0 :
+             phr[1].position > phr[0].position ? 
+             phr[1].position - phr[0].position :
+             phr[0].position - phr[1].position;
+  }
+  
+  UdmSort((void*)phr, num, sizeof(UDM_PHR_DAT), (qsort_cmp)cmpphr);
+  for (i = 1; i < num-1; i++)
+  {
+    if (phr[i-1].order == phr[i].order)
     {
-      sum += phr[i].position - phr[i-1].position;
-      cur_order = phr[i].order;
-      np++;
+      if (phr[i].order == phr[i+1].order)
+      {
+        /* w1 w1 w1 */
+      }
+      else
+      {
+        /* w1 w1 w2 */
+      }
+    }
+    else
+    {
+      if (phr[i].order == phr[i+1].order)
+      {
+        /* w1 w2 w2 */
+      }
+      else
+      {
+        /* w1 w2 w1 */
+        size_t diff1= phr[i].position - phr[i-1].position;
+        size_t diff2= phr[i+1].position - phr[i].position;
+        sum+= diff1 < diff2 ? diff1 : diff2;
+        np++;
+      }
     }
   }
   return (np) ? (sum / np) : 0;
@@ -1064,6 +1094,9 @@ static void UdmGroupByURLNewRel(UDM_RESULT *Res)
 }
 
 
+#define NWF_SIZE 256
+#define NWF_FACTOR(nwf, n) (nwf[n >= NWF_SIZE ? NWF_SIZE - 1 : n])
+
 void UdmGroupByURL(UDM_AGENT *query,UDM_RESULT *Res)
 {
   UDM_STACK_ITEM *temp_items;
@@ -1081,7 +1114,8 @@ void UdmGroupByURL(UDM_AGENT *query,UDM_RESULT *Res)
   UDM_URL_CRD *CrdTo= Res->CoordList.Coords;
   UDM_URL_CRD *CrdFrom= Res->CoordList.Coords + 1;
   UDM_URL_CRD *CrdLast= Res->CoordList.Coords + Res->CoordList.ncoords;
-  float Rsum_factor;
+  float Rsum_factor, nword_factor;
+  float nwf[NWF_SIZE];
   
   if(!Res->CoordList.ncoords) return;
 
@@ -1093,6 +1127,10 @@ void UdmGroupByURL(UDM_AGENT *query,UDM_RESULT *Res)
   
   
   UdmWeightFactorsInit(UdmVarListFindStr(&query->Conf->Vars, "wf", ""), wf);
+  for (i= 0; i < NWF_SIZE; i++)
+  {
+    nwf[i]= 0.9 + ((float) i / Res->CoordList.ncoords)*0.1;
+  }
   phr = (UDM_PHR_DAT*)UdmMalloc(phr_m * sizeof(UDM_PHR_DAT));
   count = (size_t*)UdmMalloc(count_size);
   D_size = numcosine * sizeof(unsigned int);
@@ -1173,7 +1211,8 @@ void UdmGroupByURL(UDM_AGENT *query,UDM_RESULT *Res)
   phr[phr_n].order = Res->WWList.Word[UDM_WRDNUM(Crd[0].coord)].order;
   count[phr[phr_n].order]++;
   phr_n++;
-  D[1 + nsections * UDM_WRDNUM(Crd[0].coord) + UDM_WRDSEC(Crd[0].coord)] = wf[UDM_WRDSEC(Crd[0].coord)];
+  D[1 + nsections * UDM_WRDNUM(Crd[0].coord) + UDM_WRDSEC(Crd[0].coord)]=
+   wf[UDM_WRDSEC(Crd[0].coord)];
 
 
   for( ; CrdFrom < CrdLast; CrdFrom++)
@@ -1191,14 +1230,17 @@ void UdmGroupByURL(UDM_AGENT *query,UDM_RESULT *Res)
         phr = (UDM_PHR_DAT*)UdmRealloc(phr, phr_m * sizeof(UDM_PHR_DAT));
       }
 
-      D[1 + nsections * UDM_WRDNUM(CrdFrom->coord) + UDM_WRDSEC(CrdFrom->coord)] = 
-        (wf[UDM_WRDSEC(CrdFrom->coord)] << 2) + UdmOriginWeight(Res->WWList.Word[UDM_WRDNUM(CrdFrom->coord)].origin);
+      D[1 + nsections * UDM_WRDNUM(CrdFrom->coord) + UDM_WRDSEC(CrdFrom->coord)]= 
+        (wf[UDM_WRDSEC(CrdFrom->coord)] << 2) +
+         UdmOriginWeight(Res->WWList.Word[UDM_WRDNUM(CrdFrom->coord)].origin);
     }
     else
     {
     
       /* Next document */
 
+      nword_factor= NWF_FACTOR(nwf, phr_n);
+      
       switch(search_mode)
       {
         case UDM_MODE_BOOL:
@@ -1207,7 +1249,8 @@ void UdmGroupByURL(UDM_AGENT *query,UDM_RESULT *Res)
           if(UdmCalcBoolItems(temp_items, nitems, count))
           {
             D[0]= CalcAverageWordDistance(phr, phr_n);
-            CrdTo->coord = UdmCalcCosineWeight(R, D, numcosine, Rsum_factor);
+            CrdTo->coord = UdmCalcCosineWeight(R, D, numcosine, 
+                                               Rsum_factor, nword_factor);
             CrdTo++;
           }
           break;
@@ -1217,7 +1260,8 @@ void UdmGroupByURL(UDM_AGENT *query,UDM_RESULT *Res)
           if (z < Res->WWList.nuniq && count[z] == 0) break;
         default:
           D[0] = CalcAverageWordDistance(phr, phr_n);
-          CrdTo->coord = UdmCalcCosineWeight(R, D, numcosine, Rsum_factor);
+          CrdTo->coord = UdmCalcCosineWeight(R, D, numcosine,
+                                             Rsum_factor, nword_factor);
           CrdTo++;
       }
 
@@ -1231,15 +1275,18 @@ void UdmGroupByURL(UDM_AGENT *query,UDM_RESULT *Res)
       count[phr[phr_n].order]++;
       phr_n++;
 
-      D[1 + nsections * UDM_WRDNUM(CrdFrom->coord) + UDM_WRDSEC(CrdFrom->coord)] = 
-        (wf[UDM_WRDSEC(CrdFrom->coord)] << 2) + UdmOriginWeight(Res->WWList.Word[UDM_WRDNUM(CrdFrom->coord)].origin);
+      D[1 + nsections * UDM_WRDNUM(CrdFrom->coord) + UDM_WRDSEC(CrdFrom->coord)]= 
+        (wf[UDM_WRDSEC(CrdFrom->coord)] << 2) +
+        UdmOriginWeight(Res->WWList.Word[UDM_WRDNUM(CrdFrom->coord)].origin);
     }
   }
 
   /* Check last word */
 
   D[0] = CalcAverageWordDistance(phr, phr_n);
-  CrdTo->coord = UdmCalcCosineWeight(R, D, numcosine, Rsum_factor);
+  nword_factor= NWF_FACTOR(nwf, phr_n);
+  CrdTo->coord = UdmCalcCosineWeight(R, D, numcosine,
+                                     Rsum_factor, nword_factor);
 
   switch(search_mode)
   {
@@ -1258,6 +1305,7 @@ void UdmGroupByURL(UDM_AGENT *query,UDM_RESULT *Res)
       if (CrdTo->coord)
         CrdTo++;
   }
+
   Res->CoordList.ncoords= CrdTo - Crd;
 
 
@@ -1409,9 +1457,9 @@ int UdmParseQueryString(UDM_AGENT * Agent,
       val=empty;
     }
     UdmUnescapeCGIQuery(str,val);
-    UdmVarListAddStr(vars,tok,str);
+    UdmVarListAddQueryStr(vars,tok,str);
     udm_snprintf(qname, 256, "query.%s", tok);
-    UdmVarListAddStr(vars, qname, str);
+    UdmVarListAddQueryStr(vars, qname, str);
     
     sprintf(str,"Limit-%s",tok);
     if((lim=UdmVarListFindStr(vars,str,NULL)))
@@ -1465,113 +1513,149 @@ int UdmParseQueryString(UDM_AGENT * Agent,
 }
 
 
+
+static int
+UdmWordInWWList(UDM_WIDEWORDLIST *List, int *tok, size_t flen, int hlstop)
+{
+  size_t uw;
+  
+  for(uw=0; uw < List->nwords; uw++)
+  {
+    size_t slen;
+    UDM_WIDEWORD *W= &List->Word[uw];
+    if (!hlstop && W->origin == UDM_WORD_ORIGIN_STOP)
+      continue;
+    slen= W->ulen;
+    if (flen < slen)
+      continue;
+    if (flen == slen && !UdmUniStrNCaseCmp(tok, W->uword, slen))
+      return 1;
+      
+    if (flen > slen) switch (List->wm)
+    {
+      case UDM_MATCH_BEGIN:
+        if (!UdmUniStrNCaseCmp(tok, W->uword, slen))
+          return 1;
+        break;
+      case UDM_MATCH_END:
+        if (!UdmUniStrNCaseCmp(tok + flen - slen, W->uword, slen))
+          return 1;
+        break;
+      case UDM_MATCH_SUBSTR:
+        {
+          size_t l1, l2;
+          for (l1 = 0; l1 < flen; l1++)
+          {
+            if (l1 + slen > flen) break;
+            for (l2 = 0; l2 < slen; l2++)
+            {
+              if (UdmUniToLower(tok[l1 + l2]) != UdmUniToLower(W->uword[l2]))
+                break;
+            }
+            if (l2 == slen)
+            {
+              return 1;
+              break;
+            }
+          }
+        }
+        break;
+    }
+  }
+  return 0;
+}
+
+
+/* Returns a 0-terminated string */
+
+static char *
+UdmHlConvertExtWithConv(UDM_WIDEWORDLIST *List, const char *src,
+                        UDM_CONV *lc_uni, UDM_CONV *uni_bc,
+                        int hilight_stopwords)
+{
+  int		*tok, *lt, ctype, *uni;
+  int           i0= 0, i2= 2, i3= 3;
+  char          *dst;
+  size_t        srclen, dstlen= 0, fndlen= 0;
+  size_t        dstmaxlen, unimaxlen;
+  UDM_CHARSET   *sys_int= &udm_charset_sys_int;
+
+  if(!src)return NULL;
+  if ((srclen = strlen(src)) == 0) return NULL;
+  
+  dstmaxlen= srclen * 14 + 10;
+  dst= (char*)UdmMalloc(dstmaxlen);
+
+  /* Convert to unicode */
+  unimaxlen= (srclen + 10) * sizeof(int);
+  uni= (int *)UdmMalloc(unimaxlen);
+  UdmConv(lc_uni,(char*)uni, unimaxlen, src, srclen + 1);
+
+  /* Parse unicode string */
+  for (tok= UdmUniGetSepToken(uni, &lt, &ctype) ; tok ;
+       tok= UdmUniGetSepToken(NULL, &lt, &ctype))
+  {
+    int found= 0;
+    size_t flen= lt - tok;
+
+    /* Check that it is word to be marked */
+    if (ctype != UDM_UNI_SEPAR && List != NULL)
+      found= UdmWordInWWList(List, tok, flen, hilight_stopwords);
+    
+    if (found)
+    {
+      dstlen+= UdmConv(uni_bc, dst + dstlen, dstmaxlen, (char*) &i2, sizeof(i2));
+      fndlen+= flen;
+    }
+    if (uni_bc->to == sys_int)
+    {
+      memcpy(dst + dstlen, tok, sizeof(*tok) * flen);
+      dstlen+= sizeof(*tok) * flen;
+    }
+    else
+      dstlen+= UdmConv(uni_bc, dst + dstlen, dstmaxlen, (char*) tok, sizeof(*tok) * flen);
+    
+    if (found)
+      dstlen+= UdmConv(uni_bc, dst + dstlen, dstmaxlen, (char*) &i3, sizeof(i3));
+  }
+
+  dstlen+= UdmConv(uni_bc, dst + dstlen, dstmaxlen, (char*) &i0, sizeof(i0));
+
+  UdmFree(uni);
+  return dst;
+
+}
+
+
+char * UdmHlConvertExt(UDM_WIDEWORDLIST *List,const char * src,
+                       UDM_CHARSET * lcs, UDM_CHARSET * bcs, int hlstop)
+{
+  UDM_CONV lc_uni, uni_bc;
+  UdmConvInit(&lc_uni, lcs, &udm_charset_sys_int, UDM_RECODE_HTML);
+  UdmConvInit(&uni_bc, &udm_charset_sys_int, bcs, UDM_RECODE_HTML);
+  return UdmHlConvertExtWithConv(List, src, &lc_uni, &uni_bc, hlstop);
+}
+
+
+/* For PHP module compatibility */
 char * UdmHlConvert(UDM_WIDEWORDLIST *List,const char * src,
                     UDM_CHARSET * lcs, UDM_CHARSET * bcs)
 {
-  int		*tok, *lt, ctype, *uni;
-  char		*hpart;
-  size_t	len;
-  UDM_CONV	lc_uni, uni_bc;
-  UDM_CHARSET	*sys_int;
-  UDM_DSTR      buf;
-
-  if(!src)return NULL;
-
-  if ((len = strlen(src)) == 0) return NULL;
-  hpart = (char*)UdmMalloc(len * 14 + 10);
-  UdmDSTRInit(&buf, 4096);
-
-  sys_int= &udm_charset_sys_int;
-  UdmConvInit(&lc_uni, lcs, sys_int, UDM_RECODE_HTML);
-  UdmConvInit(&uni_bc,sys_int,bcs,UDM_RECODE_HTML);
-
-  /* Convert to unicode */
-  uni = (int *)UdmMalloc((len + 10) * sizeof(int));
-  UdmConv(&lc_uni,(char*)uni,sizeof(uni[0])*(len+10),src,len+1);
-
-  /* Parse unicode string */
-  tok = UdmUniGetSepToken(uni, &lt, &ctype);
-  while(tok)
-  {
-    int found=0;
-    size_t slen,flen;
-    int euchar;
-    size_t uw;
-
-    flen=lt-tok;
-
-    /* Convert token to BrowserCharset */
-    euchar=tok[flen];
-    tok[flen]=0;
-    hpart[0]='\0';
-
-    UdmConv(&uni_bc, hpart, len * 14 + 10, (char*)tok, sizeof(*tok) * flen);
-
-    tok[flen]=euchar;
-
-    /* Check that it is word to be marked */
-    if (List != NULL)
-    {
-      for(uw=0;uw<List->nwords;uw++)
-      {
-        slen = List->Word[uw].ulen;
-        if (flen < slen) continue;
-        else if (flen == slen && !UdmUniStrNCaseCmp(tok, List->Word[uw].uword, slen)) found = 1;
-        else if (flen > slen) switch (List->wm)
-        {
-          case UDM_MATCH_BEGIN:
-            if (!UdmUniStrNCaseCmp(tok, List->Word[uw].uword, slen)) found = 1;
-            break;
-          case UDM_MATCH_END:
-            if (!UdmUniStrNCaseCmp(tok + flen - slen, List->Word[uw].uword, slen)) found = 1;
-            break;
-          case UDM_MATCH_SUBSTR:
-            {
-              size_t l1, l2;
-              for (l1 = 0; l1 < flen; l1++)
-              {
-                if (l1 + slen > flen) break;
-                for (l2 = 0; l2 < slen; l2++)
-                {
-                  if (UdmUniToLower(tok[l1 + l2]) != UdmUniToLower(List->Word[uw].uword[l2])) break;
-                }
-                if (l2 == slen)
-                {
-                  found = 1;
-                  break;
-                }
-              }
-            }
-            break;
-        }
-        if (found) break;
-      }
-    }
-    
-    if(found)UdmDSTRAppend(&buf, "\2", 1);
-    UdmDSTRAppendSTR(&buf, hpart);
-    if(found)UdmDSTRAppend(&buf, "\3", 1);
-    
-    tok = UdmUniGetSepToken(NULL, &lt, &ctype);
-  }
-
-  UDM_FREE(hpart);
-  UDM_FREE(uni);
-  /*
-  fprintf(stderr,"otxt='%s'\n",src);
-  fprintf(stderr,"htxt='%s'\n",htxt);
-  */
-  return(buf.data);
+  return UdmHlConvertExt(List, src, lcs, bcs, 1);
 }
 
+
 int UdmConvert(UDM_ENV *Conf, UDM_RESULT *Res,
                UDM_CHARSET *lcs, UDM_CHARSET *bcs)
 {
   size_t i;
-  UDM_CONV lc_bc;
+  UDM_CONV lc_bc, lc_uni, uni_bc;
+  int hlstop= UdmVarListFindBool(&Conf->Vars, "ExcerptStopword", 1);
 
   /* Init converters */
   UdmConvInit(&lc_bc,lcs,bcs,UDM_RECODE_HTML);
+  UdmConvInit(&lc_uni, lcs, &udm_charset_sys_int, UDM_RECODE_HTML);
+  UdmConvInit(&uni_bc, &udm_charset_sys_int, bcs, UDM_RECODE_HTML);
   
   /* Convert word list */
   for(i=0;i<Res->WWList.nwords;i++)
@@ -1605,9 +1689,12 @@ int UdmConvert(UDM_ENV *Conf, UDM_RESULT *Res,
          - HTML with &<>" printed as is, search word hilighted.
       */
       
-      if (strcasecmp(Var->name,"URL") && strcasecmp(Var->name,"CachedCopy"))
+      if (strcasecmp(Var->name,"URL") &&
+          strcasecmp(Var->name,"CachedCopy") &&
+          strcasecmp(Var->name,"Content-Type"))
       {
-        char *newval= UdmHlConvert(&Res->WWList, Var->val, lcs, bcs);
+        char *newval= UdmHlConvertExtWithConv(&Res->WWList, Var->val,
+                                              &lc_uni, &uni_bc, hlstop);
         UDM_FREE(Var->val);
         Var->val= newval;
       }
@@ -1618,12 +1705,17 @@ int UdmConvert(UDM_ENV *Conf, UDM_RESULT *Res,
   for (i = 0; i < Conf->Vars.nvars; i++)
   {
     UDM_VAR *Var= &Conf->Vars.Var[i];
-    size_t len= strlen(Var->val);
-    char *newval= (char*)UdmMalloc(len * 12 + 1);
+    if (UdmVarType(Var) == UDM_VAR_STR &&
+        strcasecmp(Var->name, "HlBeg") &&
+        strcasecmp(Var->name, "HlEnd"))
+    {
+      size_t len= strlen(Var->val);
+      char *newval= (char*)UdmMalloc(len * 12 + 1);
     
-    UdmConv(&lc_bc, newval, len * 12 + 1, Var->val, len + 1);
-    UDM_FREE(Var->val);
-    Var->val= newval;
+      UdmConv(&lc_bc, newval, len * 12 + 1, Var->val, len + 1);
+      UDM_FREE(Var->val);
+      Var->val= newval;
+    }
   }
   
   return UDM_OK;
@@ -1844,3 +1936,105 @@ int *UdmUniSegment(UDM_AGENT *Indexer, int *ustr, const char *lang)
 
   return ustr;
 }
+
+
+int UdmResWordInfo(UDM_ENV *Env, UDM_RESULT *Res)
+{
+  size_t  len, i, j, wsize;
+  char  *wordinfo= NULL, *end;
+  size_t  corder= (size_t)-1, ccount= 0;
+  int have_suggestions= 0;
+  
+  for(len = i = 0; i < Res->WWList.nwords; i++) 
+    len += Res->WWList.Word[i].len + 64;
+  
+  wsize= (1+len)*sizeof(char);
+  wordinfo= (char*) UdmMalloc(wsize);
+  *wordinfo= '\0';
+  
+  for(i = 0; i < Res->WWList.nwords; i++)
+  {
+    if (Res->WWList.Word[i].origin == UDM_WORD_ORIGIN_QUERY ||
+        Res->WWList.Word[i].origin == UDM_WORD_ORIGIN_SPELL ||
+        Res->WWList.Word[i].origin == UDM_WORD_ORIGIN_SYNONYM)
+    {
+      if(wordinfo[0])
+        strcat(wordinfo,", ");
+      sprintf(UDM_STREND(wordinfo)," %s : %d", Res->WWList.Word[i].word, Res->WWList.Word[i].count);
+    }
+    else if (Res->WWList.Word[i].origin == UDM_WORD_ORIGIN_STOP)
+    {
+      if(wordinfo[0])
+        strcat(wordinfo,", ");
+      sprintf(UDM_STREND(wordinfo)," %s : stopword", Res->WWList.Word[i].word);
+    }
+  }
+  
+  UdmVarListReplaceStr(&Env->Vars, "WE", wordinfo);
+  
+  *wordinfo = '\0';
+  for(i = 0; i < Res->WWList.nwords; i++)
+  {
+    corder= Res->WWList.Word[i].order;
+    ccount= 0;
+    for(j= 0; j < Res->WWList.nwords; j++)
+      if (Res->WWList.Word[j].order == corder)
+        ccount += Res->WWList.Word[j].count;
+    if (Res->WWList.Word[i].origin == UDM_WORD_ORIGIN_STOP)
+    {
+      sprintf(UDM_STREND(wordinfo),"%s%s : stopword", (*wordinfo) ? ", " : "",  Res->WWList.Word[i].word);
+    }
+    else if (Res->WWList.Word[i].origin == UDM_WORD_ORIGIN_QUERY)
+    {
+      sprintf(UDM_STREND(wordinfo),"%s%s : %d / %d", 
+        (*wordinfo) ? ", " : "", Res->WWList.Word[i].word, Res->WWList.Word[i].count, ccount);
+    }
+  }
+  UdmVarListReplaceStr(&Env->Vars, "W", wordinfo);
+  
+  *wordinfo= '\0';
+  end= wordinfo;
+  for (i= 0; i < Res->WWList.nwords; i++)
+  {
+    UDM_WIDEWORD *Wi= &Res->WWList.Word[i];
+    UDM_WIDEWORD *Wb= NULL;
+   
+    if (Wi->origin == UDM_WORD_ORIGIN_QUERY)
+    {
+      if (Wi->count > 0)
+      {
+        Wb= Wi;
+      }
+      else
+      {
+        ccount= 0;
+        for (j= 0; j < Res->WWList.nwords; j++)
+        {
+          UDM_WIDEWORD *Wj= &Res->WWList.Word[j];
+          if (Wj->origin == UDM_WORD_ORIGIN_SUGGEST &&
+              Wj->order == Wi->order && Wj->count > ccount)
+          {
+            ccount= Res->WWList.Word[j].count;
+            Wb= Wj;
+            have_suggestions= 1;
+          }
+        }
+      }
+    }
+    else if (Wi->origin == UDM_WORD_ORIGIN_STOP)
+    {
+      Wb= Wi;
+    }
+    
+    if (Wb)
+    {
+      sprintf(end, "%s%s", wordinfo[0] ? " " : "", Wb->word);
+      end= end + strlen(end);
+    }
+  }
+  
+  if (have_suggestions)
+    UdmVarListReplaceStr(&Env->Vars, "WS", wordinfo);
+  UDM_FREE(wordinfo);
+  return UDM_OK;
+}
