@@ -71,6 +71,12 @@ static int UdmPgSQLQuery(UDM_DB *db, UDM_SQLRES *res, const char *q)
   PGresult *PGres;
   PGconn *pgsql;
   
+  if (res)
+  {
+    bzero((void*) res, sizeof(UDM_SQLRES));
+    res->db= db;
+  }
+
   db->errcode=0;
   if(!db->connected)
   {
@@ -79,12 +85,6 @@ static int UdmPgSQLQuery(UDM_DB *db, UDM_SQLRES *res, const char *q)
       return UDM_ERROR;
   }
   
-  if (res)
-  {
-    bzero((void*) res, sizeof(UDM_SQLRES));
-    res->db= db;
-  }
-
   pgsql= (PGconn*) db->specific;
   
   if(!(PGres= PQexec(pgsql,q)))
@@ -104,7 +104,7 @@ static int UdmPgSQLQuery(UDM_DB *db, UDM_SQLRES *res, const char *q)
   {
     PQclear(PGres);
     sprintf(db->errstr, "%s", PQerrorMessage(pgsql));
-    if(strstr(db->errstr,"duplicate") ||
+    if(strstr(db->errstr,"uplicate") ||
        strstr(db->errstr, "") ||
        strstr(db->errstr, "повторный") ||
        strcasestr(db->errstr,"Duplizierter"))
