@@ -365,6 +365,12 @@ static int UdmODBCQuery(UDM_DB *db,UDM_SQLRES *res, const char *qbuf)
 {
   int rc;
    
+  if (res)
+  {
+    bzero((void*) res, sizeof(UDM_SQLRES));
+    res->db= db;
+  }
+
   if(!db->connected)
   {
     UdmODBCInitDB(db);
@@ -379,12 +385,6 @@ static int UdmODBCQuery(UDM_DB *db,UDM_SQLRES *res, const char *qbuf)
     }
   }
 
-  if (res)
-  {
-    bzero((void*) res, sizeof(UDM_SQLRES));
-    res->db= db;
-  }
-
   rc= execDB(db,res,qbuf);
   if((db->errcode))
   {
@@ -1344,11 +1344,9 @@ static int UdmSQLiteQuery(UDM_DB *db, UDM_SQLRES *res, const char *q)
   
 #ifdef WIN32
   /** Remove heading "/" from DBName (in case of absolute path) */
-  char * tmp_dbname;
-
   if(db->DBName[2] == ':')
   {
-    tmp_dbname = UdmStrdup(db->DBName + 1);
+    char *tmp_dbname = UdmStrdup(db->DBName + 1);
     UDM_FREE(db->DBName);
     db->DBName = tmp_dbname;
   }
