@@ -45,36 +45,47 @@
 #include "udm_mutex.h"
 #include "udm_doc.h"
 #include "udm_db.h"
+#include "udm_guesser.h"
+
 
 static int UdmUniNSpace(int c)
 {
-  if (c == 0x0020) return 0;
-  if (c == 0x000D) return 0;
-  if (c == 0x000A) return 0;
-  if (c == 0x0009) return 0;
-  if (c == 0x1680) return 0;
-  if ((c >= 0x2000) && (c <= 0x200B)) return 0;
-  if (c == 0x202F) return 0;
-  if (c == 0x3000) return 0;
+  if (c > 0x0020)
+  {
+    if (c < 0x1680)
+      return 1;
+    if (c == 0x1680) return 0;
+    if ((c >= 0x2000) && (c <= 0x200B)) return 0;
+    if (c == 0x202F) return 0;
+    if (c == 0x3000) return 0;
+  }
+  else
+  {
+    if (c == 0x0020) return 0;
+    if (c == 0x000D) return 0;
+    if (c == 0x000A) return 0;
+    if (c == 0x0009) return 0;
+  }
+  
   return 1;
 }
 
+
 static int *UdmGetExcerptSourceCachedCopy (UDM_AGENT *Agent,
                                            UDM_RESULT *Res, UDM_DOCUMENT *Doc)
 {
 #ifdef HAVE_ZLIB
-  int *res;
-  size_t l, ul;
-  char *Source;
-  char *in_buf;
+  size_t l;
+  char *Source, *in_buf;
   z_stream zstream;
   UDM_DSTR buf;
-  UDM_CHARSET *sys_int = &udm_charset_sys_int;
-  UDM_CHARSET *dcs = UdmGetCharSet(UdmVarListFindStr(&Doc->Sections, "Charset", "iso-8859-1"));
-  const char *CachedCopy = UdmVarListFindStr(&Doc->Sections, "CachedCopy", NULL);
+  UDM_CHARSET *sys_int= &udm_charset_sys_int;
+  UDM_CHARSET *dcs= UdmGetCharSet(UdmVarListFindStr(&Doc->Sections, "Charset", "iso-8859-1"));
+  const char *CachedCopy= UdmVarListFindStr(&Doc->Sections, "CachedCopy", NULL);
+  const char *content_type= UdmVarListFindStr(&Doc->Sections, "Content-Type", NULL);
   UDM_HTMLTOK tag;
   const char *htok, *last;
-  UDM_CONV conv;
+  int hlstop= UdmVarListFindBool(&Agent->Conf->Vars, "ExcerptStopword", 1);
   
   if (!CachedCopy || !sys_int || !dcs)
     return NULL;
@@ -107,41 +118,51 @@ static int *UdmGetExcerptSourceCachedCopy (UDM_AGENT *Agent,
   inflateEnd(&zstream);
   Source[zstream.total_out] = 0;
   UdmFree(in_buf);
-  
+
+  if (!content_type ||
+      (strcmp(content_type, "text/plain") &&
+       strcmp(content_type, "text/xml") &&
+       strcmp(content_type, "text/html")))
+    content_type= UdmGuessContentType(Source, zstream.total_out);
   UdmDSTRInit(&buf, 1024);
-  UdmHTMLTOKInit(&tag);
-  htok= UdmHTMLToken(Source, &last, &tag);
-  do
+
+  if (!strcmp(content_type, "text/xml"))
   {
-    if (tag.type == UDM_HTML_TXT && ! tag.script && ! tag.comment && ! tag.style)
-      UdmDSTRAppend(&buf, htok, last - htok);
-    else
-      UdmDSTRAppend(&buf, " ", 1);
-  } while ((htok = UdmHTMLToken(NULL, &last, &tag)));
-  UdmFree(Source);
+    UdmHTMLTOKInit(&tag);
+    htok= UdmHTMLToken(Source, &last, &tag);
+    do
+    {
+      if (tag.type == UDM_HTML_TXT &&
+          !tag.script && !tag.comment && ! tag.style)
+        UdmDSTRAppend(&buf, htok, last - htok);
+      else
+        UdmDSTRAppend(&buf, " ", 1);
+    } while ((htok = UdmHTMLToken(NULL, &last, &tag)));
+  }
+  else if (!strcmp(content_type, "text/html"))
+  {
+    UdmHTMLTOKInit(&tag);
+    htok= UdmHTMLToken(Source, &last, &tag);
+    do
+    {
+      if (tag.type == UDM_HTML_TXT && tag.body &&
+          !tag.script && !tag.comment && ! tag.style)
+        UdmDSTRAppend(&buf, htok, last - htok);
+      else
+        UdmDSTRAppend(&buf, " ", 1);
+    } while ((htok = UdmHTMLToken(NULL, &last, &tag)));
+  }
+  else
+  {
+    UdmDSTRAppend(&buf, Source, zstream.total_out);
+  }
   
-  Source= UdmHlConvert(&Res->WWList, buf.data, dcs, dcs);
+  UdmFree(Source);
+
+  Source= UdmHlConvertExt(&Res->WWList, buf.data, dcs, sys_int, hlstop);
   UdmDSTRFree(&buf);
-  if (! Source)
-    return NULL;
-  l= strlen(Source);
-  
-  if (!(res= UdmMalloc(sizeof(int) * (l + 1))))
-  {
-    UdmFree(Source);
-    return(NULL);
-  }
-  
-  UdmConvInit(&conv, dcs, sys_int, UDM_RECODE_HTML);
-  ul= UdmConv(&conv, (char *)res, sizeof(int) * (l + 1), Source, l);
-  UdmFree(Source);
-  if (ul < 0)
-  {
-    UdmFree(res);
-    return(NULL);
-  }
-  res[ul / sizeof(int)] = 0;
-  return res;
+  return (int*) Source;
+
 #else
   return NULL;
 #endif
@@ -190,13 +211,13 @@ __C_LINK char * __UDMCALL UdmExcerptDoc (UDM_AGENT *Agent,
   
   if (!sys_int || !cs)
     return NULL;
-  
+
   if (!(Source= UdmGetExcerptSourceCachedCopy(Agent, Res, Doc)) &&
       !(Source= UdmGetExcerptSourceBody(Agent, Res, Doc)))
     return(NULL);
-  
+
   ul= UdmUniLen(Source);
-  
+
   /* Strip whitespaces */
   for (i= 0, j= 0; i < ul; i++)
   {
@@ -207,7 +228,7 @@ __C_LINK char * __UDMCALL UdmExcerptDoc (UDM_AGENT *Agent,
   }
   Source[j] = 0;
   ul = j;
-  
+
   /* Get excerpt */
   UdmDSTRInit(&buf, 1024);
   for (i= 0; i < ul; i++)
@@ -254,6 +275,7 @@ __C_LINK char * __UDMCALL UdmExcerptDoc (UDM_AGENT *Agent,
     UdmDSTRFree(&buf);
     return NULL;
   }
+
   UdmConvInit(&conv, sys_int, cs, UDM_RECODE_HTML);
   l= UdmConv(&conv, res, ul, buf.data, buf.size_data);
   UdmDSTRFree(&buf);
